import pytest

from syftr.configuration import cfg
from syftr.studies import StudyConfig
from tests.check_trials import check_trials_are_flawless
from tests.e2e.optimization import run_and_test_optimization


@pytest.mark.parametrize(
    "config_file_name",
    [
        "test-tools-anthropic-haiku-35.yaml",
        # "test-tools-gemini-flash2.yaml", # Seems to be flaky with tools
        "test-tools-gpt-4o-mini.yaml",
        # "test-tools-llama-33-70B.yaml", # Seems to be flaky with tools
        "test-tools-mistral-large.yaml",
        "test-tools-o3-mini.yaml",
        "test-tools-phi-4.yaml",
    ],
)
def test_agents_with_tool_use(config_file_name):
    study_config_file = cfg.paths.test_studies_dir / config_file_name
    study_config = StudyConfig.from_file(study_config_file)
    study_config.optimization.use_hf_embedding_models = False

    df_trials = run_and_test_optimization(study_config)
    check_trials_are_flawless(df_trials)
    assert len(df_trials) == 5, f"Expected 5 trials, but got {len(df_trials)}"



import json

import numpy as np
import pandas as pd
from paretoset import paretoset
from sklearn.cluster import KMeans

from syftr.configuration import cfg
from syftr.huggingface_helper import get_embedding_model
from syftr.logger import logger
from syftr.optuna_helper import get_completed_trials
from syftr.studies import StudyConfig


def get_n_pareto_fronts(df_tmp, max_n: int = 1) -> pd.DataFrame:
    assert max_n > 0, "Wrong number of Pareto fronts"
    dfs = []
    df_tmp = df_tmp.copy()
    df_tmp["front"] = 1
    mask = paretoset(df_tmp[["values_0", "values_1"]], sense=["max", "min"])
    dfs.append(df_tmp[mask])
    if max_n > 1:
        df_next = df_tmp[~mask]
        df_pareto = get_n_pareto_fronts(df_next, max_n - 1)
        df_pareto["front"] += 1
        dfs.append(df_pareto)
    return pd.concat(dfs, ignore_index=True)


def get_top_performing_trials(
    df_trials: pd.DataFrame, max_pareto_fronts: int = 1
) -> pd.DataFrame:
    logger.info("Getting top-performing flows")
    df_result = df_trials.groupby("study_name").apply(
        lambda x: get_n_pareto_fronts(x, max_pareto_fronts)
    )
    df_result = df_result.groupby("user_attrs_flow", as_index=False).first()
    return df_result.sort_index()


def get_selected_trials(
    df_top: pd.DataFrame, embedding_model: str, max_total: int
) -> pd.DataFrame:
    logger.info(f"Clustering flows for selection using '{embedding_model}'")
    embedder, _ = get_embedding_model(embedding_model)
    assert embedder, "No embedding model"
    flows = list(df_top["user_attrs_flow"].values)
    embeddings = np.array([embedder.get_query_embedding(text) for text in flows])
    kmeans = KMeans(n_clusters=max_total, random_state=0)
    df_top["cluster"] = kmeans.fit_predict(embeddings)
    return df_top.loc[df_top.groupby("cluster")["front"].idxmin()].sort_index()


def get_examples(study_config: StudyConfig):
    study_names = study_config.transfer_learning.studies
    assert study_names, "No studies for transfer learning"
    max_fronts = study_config.transfer_learning.max_fronts
    max_total = study_config.transfer_learning.max_total
    success_rate = study_config.transfer_learning.success_rate

    storage = cfg.database.get_optuna_storage()
    df_trials = get_completed_trials(
        study=study_names, storage=storage, success_rate=success_rate
    )
    df_top = get_top_performing_trials(df_trials, max_fronts)

    if len(df_top) <= max_total:
        return [json.loads(flow) for flow in df_top["user_attrs_flow"]]

    embedding_model = study_config.transfer_learning.embedding_model
    df_selected = get_selected_trials(df_top, embedding_model, max_total)
    return [json.loads(flow) for flow in df_selected["user_attrs_flow"]]



import hashlib
import typing as T
from abc import ABC, abstractmethod
from functools import cached_property
from pathlib import Path

import datasets
import pandas as pd
from datasets import Features, Sequence, Value
from llama_index.core import Document, SimpleDirectoryReader
from llama_index.core.evaluation.correctness import DEFAULT_USER_TEMPLATE
from overrides import overrides
from pydantic import BaseModel, field_validator
from ray._private.utils import get_ray_temp_dir

from syftr.configuration import cfg
from syftr.core import QAPair, QARTriplet
from syftr.utils.locks import distributed_lock


class PartitionMap(BaseModel):
    sample: str = "sample"
    train: str = "train"
    test: str = "test"
    holdout: str = "holdout"


class SyftrQADataset(BaseModel, ABC):
    """Container and utilities for dataset with remote storage.

    Instances of this class do _not_ store the dataset; only pointers to it.
    Therefore, this class can be safely passed to Ray tune.
    """

    # Canonical name for the dataset
    # Set this to a unique string Literal on your subclass
    # Must be present when loading a StudyConfig from a yaml
    xname: T.Literal["SyftrQADataset"] = "SyftrQADataset"

    # Partition names for this dataset as it is stored on disk
    storage_partitions: T.List[str] = ["sample", "train", "test", "holdout"]
    # How to map requested partition to the storage partitions
    # eg. MyDataset(partition_map={'test': 'sample'}) to run on the sample partition
    partition_map: PartitionMap = PartitionMap()

    # timeouts
    load_examples_timeout_s: int = 3600
    load_grounding_data_timeout_s: int = 3600

    # The evaluation user template specifies content and format of the information that is passed to the evaluator.
    # The template must include Python string templates for the following fields:
    #     'query', 'reference_answer', and 'generated_answer'.
    # The user may use this template to insert task-specific instructions to the evaluator to ensure the appropriate
    # criteria are being used.
    eval_user_template: str = DEFAULT_USER_TEMPLATE

    @property
    def name(self) -> str:
        """Subclasses may dynamically construct name."""
        return self.xname

    def _get_storage_partition(self, canonical_partition: str) -> str:
        return getattr(self.partition_map, canonical_partition)

    @abstractmethod
    def iter_examples(self, partition="test") -> T.Iterator[QAPair]:
        pass

    @abstractmethod
    def iter_grounding_data(self, partition="test") -> T.Iterator[Document]:
        pass


class InfiniteBenchHF(SyftrQADataset):
    xname: T.Literal["infinitebench_hf"] = "infinitebench_hf"  # type: ignore
    subset: T.Literal["longbook_qa_eng"] = "longbook_qa_eng"
    description: str = "The dataset contains a large number of books."

    def _load_raw_dataset(self) -> pd.DataFrame:
        ft = Features(
            {
                "id": Value("int64"),
                "context": Value("string"),
                "input": Value("string"),
                "answer": Sequence(Value("string")),
                "options": Sequence(Value("string")),
            }
        )

        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "xinrongzhang2022/InfiniteBench",
                features=ft,
                split=self.subset,
                cache_dir=cfg.paths.huggingface_cache,
            )
        df = dataset.to_pandas()
        return df

    def _add_partitions(self, df: pd.DataFrame) -> pd.DataFrame:
        df["book_id"] = df.context.factorize()[0]
        books = df.book_id.unique()
        book_partitions = {
            "sample": books[:1],
            "train": books[1:23],
            "test": books[23:46],
            "holdout": books[46:69],
        }

        def label_partition(book_id: int) -> str:
            for partition, book_range in book_partitions.items():
                if book_id in book_range:
                    return partition
            raise IndexError(f"Book id {book_id} is out of range of {book_partitions=}")

        df["partition"] = df.book_id.apply(label_partition)
        return df

    @property
    def _dataset(self) -> pd.DataFrame:
        df = self._load_raw_dataset()
        df = self._add_partitions(df)
        return df

    def _row_to_qapair(self, row: T.Dict[str, T.Any]) -> QAPair:
        """Dataset-specific conversion of row to QAPair struct.

        Invoked by iter_examples.

        Default implementation assumes row is already in QAPair format.
        """
        return QAPair(
            question=row["input"],
            answer=str(row["answer"]),
            id=str(row["id"]),
            context={"book_start": row["context"][:100]},
            supporting_facts=[],
            difficulty="",
            qtype="",
        )

    @overrides
    def iter_examples(self, partition="test") -> T.Iterator[QAPair]:
        df = self._dataset
        partition = self._get_storage_partition(partition)
        df = df[df.partition == partition]
        for _, row in df.iterrows():
            yield self._row_to_qapair(row)

    @overrides
    def iter_grounding_data(self, partition="test") -> T.Iterator[Document]:
        df = self._dataset
        partition = self._get_storage_partition(partition)
        df = df[df.partition == partition]
        for book in df.context.unique():
            yield Document(
                text=book,
            )


class FinanceBenchHF(SyftrQADataset):
    xname: T.Literal["financebench_hf"] = "financebench_hf"  # type: ignore
    description: str = (
        "Financial dataset that contains everything about finance, including real-world financial documents, "
        "SEC filings, earning reports, call transcripts, and much more. "
        "It has all the financial live data, historical data, just about everything about finance, for instance, "
        "definitions and explanations of financial term, "
        "insights on company revenues, mergers, founders, or stock performance, "
        "details on financial laws, compliance, or government policies, "
        "information required to evaluated finance risk, and "
        "information about banking operations, credit systems, or loan structures."
    )

    def _load_grounding_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/financebench",
                "groundingdata",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset

    def _load_qa_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/financebench",
                "qapairs",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset

    @overrides
    def iter_grounding_data(
        self, partition="test", **load_kwargs
    ) -> T.Iterator[Document]:
        assert partition in self.storage_partitions
        grounding_dataset = self._load_grounding_dataset()
        partition = self._get_storage_partition(partition)
        for row in grounding_dataset[partition]:
            yield Document(
                text=row["html"],
                metadata={"file_name": row["filename"]},
            )

    def _row_to_qapair(self, row):
        """Dataset-specific conversion of row to QAPair struct.

        Invoked by iter_examples.

        Default implementation assumes row is already in QAPair format.
        """
        return QAPair(
            question=row["question"],
            answer=row["answer"],
            id=str(row["financebench_id"]),
            context=row["evidence"],
            supporting_facts=[row["justification"]],
            difficulty="",
            qtype=row["question_type"],
            gold_evidence=row["html_evidence"],
        )

    @overrides
    def iter_examples(self, partition="test") -> T.Iterator[QAPair]:
        partition = self._get_storage_partition(partition)
        qa_examples = self._load_qa_dataset()
        for row in qa_examples[partition]:
            yield self._row_to_qapair(row)


class SyntheticFinanceBenchHF(FinanceBenchHF):
    xname: T.Literal["synthetic_financebench_hf"] = "synthetic_financebench_hf"  # type: ignore

    def _load_qa_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/financebench",
                "qapairs_synthetic",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset

    def _row_to_qapair(self, row):
        """Dataset-specific conversion of row to QAPair struct.

        Invoked by iter_examples.

        Default implementation assumes row is already in QAPair format.
        """
        return QAPair(
            question=row["query"],
            answer=row["reference_answer"],
            id=str(row["id"]),
            context={},
            supporting_facts=[],
            difficulty="default",
            qtype="default",
        )


class HotPotQAHF(SyftrQADataset):
    xname: T.Literal["hotpotqa_hf"] = "hotpotqa_hf"  # type: ignore
    subset: str = "dev"  # train_hard, dev
    description: str = (
        "This dataset is a vast collection of all kind of information that you can find on Wikipedia. "
        "It can be used, for instance, to retrieve straightforward facts from one or more documents, "
        "compare two entities based on shared attributes, "
        "identify relationships, roles, or attributes of entities, "
        "reason about dates, timelines, or chronological order, "
        "determine geographical relationships or locations, "
        "explain causes or sequences of events or processes, "
        "synthesize facts from multiple documents to infer answers, and "
        "validate or refute premises in the context of the question."
    )

    @property
    def name(self) -> str:
        return f"{self.xname}/{self.subset}"

    def _load_grounding_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/hotpotqa",
                f"groundingdata_{self.subset}",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset

    def _load_qa_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/hotpotqa",
                f"qapairs_{self.subset}",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset

    @overrides
    def iter_grounding_data(
        self,
        partition="test",
    ) -> T.Iterator[Document]:
        assert partition in self.storage_partitions
        grounding_dataset = self._load_grounding_dataset()
        partition = self._get_storage_partition(partition)
        for row in grounding_dataset[partition]:
            yield Document(
                text=row["text"],
            )

    def _row_to_qapair(self, row):
        """Dataset-specific conversion of row to QAPair struct.

        Invoked by iter_examples.

        Default implementation assumes row is already in QAPair format.
        """
        return QAPair(
            question=row["question"],
            answer=row["answer"],
            id=str(row["id"]),
            context=[{title: sentence} for title, sentence in row["context"]],
            supporting_facts=row["supporting_facts"],
            difficulty=row["level"],
            qtype=row["type"],
            gold_evidence=[sentence for _, sentence in row["context"]],
        )

    @overrides
    def iter_examples(self, partition="test") -> T.Iterator[QAPair]:
        partition = self._get_storage_partition(partition)
        qa_examples = self._load_qa_dataset()
        for row in qa_examples[partition]:
            yield self._row_to_qapair(row)


class SyntheticHotPotQAHF(HotPotQAHF):
    xname: T.Literal["synthetic_hotpotqa_hf"] = "synthetic_hotpotqa_hf"  # type: ignore

    @property
    def name(self) -> str:
        return f"{self.xname}/{self.subset}"

    def _load_qa_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/hotpotqa",
                f"qapairs_synthetic_{self.subset}",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset

    def _row_to_qapair(self, row):
        """Dataset-specific conversion of row to QAPair struct.

        Invoked by iter_examples.

        Default implementation assumes row is already in QAPair format.
        """
        return QAPair(
            question=row["query"],
            answer=row["reference_answer"],
            id=str(row["id"]),
            context={},
            supporting_facts=[],
            difficulty="default",
            qtype="default",
        )


class CragTask3HF(SyftrQADataset):
    xname: T.Literal["crag_hf"] = "crag_hf"  # type: ignore
    subset: str = "sports"  # finance, movie, music, sports, open

    _descriptions = {
        "sports": (
            "This resource contains everything about sports, including sports news, sports events, "
            "sports statistics, sports schedules, and more. "
            "It contains historical data, live data, just about everything about sports, for instance, "
            "information about athlete achievements, teams, or career stats, "
            "information on match dates, scores, or tournaments, "
            "team performances or player statistics, "
            "key events or records in sports history, and "
            "information about rules or formats in specific sports."
        ),
        "finance": (
            "This resource contains everything about finance, including financial news, "
            "market data, financial reports, and more. "
            "It contains historical data, live data, just about everything about finance, for instance, "
            "stock prices, trends, or market indices, "
            "revenue, founders, or headquarters of companies, "
            "major financial events or policy changes, "
            "comparing returns or risks of different investments, and "
            "financial terms or concepts explained."
        ),
        "movie": (
            "This resource contains everything about movies, including movie news, reviews, "
            "box office data, and more. "
            "It contains historical data, live data, just about everything about movies, for instance, "
            "information about actors, directors, or crew roles, "
            "movies by a specific actor or director, "
            "key events or summaries from movie plots, "
            "recognition received by films or individuals, and "
            "dates or production details for movies."
        ),
        "music": (
            "This resource contains everything about music, including music news, charts, "
            "artist information, and more. "
            "It contains historical data, live data, just about everything about music, for instance, "
            "information on musicians, albums, or band members, "
            "albums or songs released by specific artists, "
            "information about song lyrics or their significance, "
            "Grammys or other recognitions for artists or albums, and "
            "details about music styles or trends over time."
        ),
        "open": (
            "This resource contains everything about open domain, including general knowledge, "
            "trivia, fun facts, and more. "
            "It contains historical data, live data, just about everything about open domain, for instance, "
            "general knowledge about diverse topics, "
            "facts that combine multiple subject areas, "
            "unique or interesting facts across domains (trivia), and "
            "further material useful for complex, unconstrained reasoning without a specific domain."
        ),
    }

    @property
    def name(self) -> str:
        return f"{self.xname}/{self.subset}"

    @property
    def description(self) -> str:
        return self._descriptions[self.subset]

    def _load_grounding_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/crag",
                f"groundingdata_{self.subset}",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset

    def _load_qa_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/crag",
                f"qapairs_{self.subset}",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset

    @overrides
    def iter_grounding_data(self, partition="test") -> T.Iterator[Document]:
        assert partition in self.storage_partitions
        grounding_dataset = self._load_grounding_dataset()
        partition = self._get_storage_partition(partition)
        for row in grounding_dataset[partition]:
            yield Document(
                text=row["markdown"],
                metadata={"file_name": row["filename"]},
            )

    def _row_to_qapair(self, row):
        """Dataset-specific conversion of row to QAPair struct.

        Invoked by iter_examples.

        Default implementation assumes row is already in QAPair format.
        """
        context = {result["_hash"]: str(result) for result in row["search_results"]}
        return QAPair(
            question=row["query"],
            answer=row["answer"],
            id=str(row["interaction_id"]),
            context=context,
            supporting_facts=[],
            difficulty="default",
            qtype=row["question_type"],
        )

    @overrides
    def iter_examples(self, partition="test") -> T.Iterator[QAPair]:
        partition = self._get_storage_partition(partition)
        qa_examples = self._load_qa_dataset()
        for row in qa_examples[partition]:
            yield self._row_to_qapair(row)


class SyntheticCragTask3HF(CragTask3HF):
    xname: T.Literal["synthetic_crag_hf"] = "synthetic_crag_hf"  # type: ignore

    @property
    def name(self) -> str:
        return f"{self.xname}/{self.subset}"

    def _load_qa_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/crag",
                f"qapairs_synthetic_{self.subset}",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset

    def _row_to_qapair(self, row):
        """Dataset-specific conversion of row to QAPair struct.

        Invoked by iter_examples.

        Default implementation assumes row is already in QAPair format.
        """
        return QAPair(
            question=row["query"],
            answer=row["reference_answer"],
            id=str(row["id"]),
            context={},
            supporting_facts=[],
            difficulty="default",
            qtype="default",
        )


class DRDocsHF(SyftrQADataset):
    xname: T.Literal["drdocs_hf"] = "drdocs_hf"  # type: ignore
    description: str = (
        "The dataset contains comprehensive information about DataRobot, including its API, documentation, examples, "
        "key features, platform architecture, integrations, setup guides, data handling, feature engineering, EDA tools, "
        "automated machine learning, model management, deployment options, monitoring, REST API, batch predictions, "
        "real-time scoring, custom recipes, retraining, lifecycle management, bias detection, explainability, diagnostics, "
        "cross-validation, leaderboard insights, time series modeling, data governance, security, user roles, Python/R usage, "
        "custom blueprints, external model integration, Docker deployments, API reference, BI tool integration, workflow automation, "
        "multimodal modeling, NLP, image recognition, hyperparameter tuning, performance optimization, resource management, "
        "parallel processing, drift detection, retraining triggers, industry use cases, tutorials, case studies, common issues, "
        "debugging tips, FAQs, support access, community resources, and release notes."
    )

    def _load_grounding_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/drdocs",
                "groundingdata",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset

    def _load_qa_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/drdocs",
                "qapairs",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset

    @overrides
    def iter_grounding_data(self, partition="notused") -> T.Iterator[Document]:
        # There is no partition. The grounding dataset is the same
        # across all partitions of the qa pairs.
        grounding_dataset = self._load_grounding_dataset()
        for row in grounding_dataset["train"]:
            yield Document(
                text=row["markdown"],
                metadata={"file_name": row["filename"]},
            )

    def _row_to_qapair(self, row):
        """Dataset-specific conversion of row to QAPair struct.

        Invoked by iter_examples.

        Default implementation assumes row is already in QAPair format.
        """
        return QAPair(
            question=row["question"],
            answer=row["answer"],
            id=str(row["id"]),
            context={},
            supporting_facts=[],
            difficulty="default",
            qtype="default",
        )

    @overrides
    def iter_examples(self, partition="test") -> T.Iterator[QAPair]:
        assert partition in self.storage_partitions
        partition = self._get_storage_partition(partition)
        qa_examples = self._load_qa_dataset()
        for row in qa_examples[partition]:
            yield self._row_to_qapair(row)


class PhantomWikiv050(SyftrQADataset):
    xname: T.Literal["phantomwikiv050_hf"] = "phantomwikiv050_hf"  # type: ignore

    # subset choices are:
    # 'depth_20_size_25_seed_1',
    # 'depth_20_size_25_seed_2',
    # 'depth_20_size_25_seed_3',
    # 'depth_20_size_50_seed_1',
    # 'depth_20_size_50_seed_2',
    # 'depth_20_size_50_seed_3',
    # 'depth_20_size_100_seed_1',
    # 'depth_20_size_100_seed_2',
    # 'depth_20_size_100_seed_3',
    # 'depth_20_size_200_seed_1',
    # 'depth_20_size_200_seed_2',
    # 'depth_20_size_200_seed_3',
    # 'depth_20_size_300_seed_1',
    # 'depth_20_size_300_seed_2',
    # 'depth_20_size_300_seed_3',
    # 'depth_20_size_400_seed_1',
    # 'depth_20_size_400_seed_2',
    # 'depth_20_size_400_seed_3',
    # 'depth_20_size_500_seed_1',
    # 'depth_20_size_500_seed_2',
    # 'depth_20_size_500_seed_3',
    # 'depth_20_size_1000_seed_1',
    # 'depth_20_size_1000_seed_2',
    # 'depth_20_size_1000_seed_3',
    # 'depth_20_size_2500_seed_1',
    # 'depth_20_size_2500_seed_2',
    # 'depth_20_size_2500_seed_3',
    # 'depth_20_size_5000_seed_1',
    # 'depth_20_size_5000_seed_2',
    # 'depth_20_size_5000_seed_3',
    # 'depth_20_size_10000_seed_1',
    # 'depth_20_size_10000_seed_2',
    # 'depth_20_size_10000_seed_3'

    subset: str = "depth_20_size_10000_seed_3"
    description: str = (
        "This dataset contains data from PhantomWiki, which "
        "is a framework for generating unique, factually"
        "and consistent document corpora with diverse question-answer pairs."
        "Unlike prior work, PhantomWiki is neither a fixed dataset, nor is it"
        "based on any existing data. Instead, a new PhantomWiki instance is "
        "generated on demand for each evaluation. PhantomWiki generates a "
        "fictional universe of characters along with a set of facts. "
        "We reflect these facts in a large-scale corpus, mimicking the "
        "style of fan-wiki websites. Then we generate question-answer pairs "
        "with tunable difficulties, encapsulating the types of multi-hop "
        "questions commonly considered in the question-answering (QA) literature."
    )

    def _get_partition_range(self, partition: str):
        partition_ranges = {
            "sample": range(0, 10),
            "train": range(0, 100),
            "test": range(100, 400),
            "holdout": range(400, 500),
        }
        if partition in partition_ranges:
            return partition_ranges[partition]

    def _load_grounding_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "kilian-group/phantom-wiki-v050",
                "text-corpus",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset[self.subset]

    def _load_qa_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "kilian-group/phantom-wiki-v050",
                "question-answer",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset[self.subset]

    @overrides
    def iter_grounding_data(self, partition="notused") -> T.Iterator[Document]:
        # There is no partition. The grounding dataset is the same
        # across all partitions of the qa pairs.
        grounding_dataset = self._load_grounding_dataset()
        for row in grounding_dataset:
            yield Document(
                text=row["article"],
                metadata={"title": row["title"]},
            )

    def _row_to_qapair(self, row):
        """Dataset-specific conversion of row to QAPair struct.

        Invoked by iter_examples.

        Default implementation assumes row is already in QAPair format.
        """
        return QAPair(
            question=row["question"],
            answer=" ".join(row["answer"]),
            id=str(row["id"]),
            context={},
            supporting_facts=[],
            difficulty=str(row["difficulty"]),
            qtype=str(row["type"]),
        )

    @overrides
    def iter_examples(self, partition="test") -> T.Iterator[QAPair]:
        assert partition in self.storage_partitions
        partition = self._get_storage_partition(partition)
        qa_examples = self._load_qa_dataset()
        partition_range = self._get_partition_range(partition)

        for i in partition_range:
            row = qa_examples[i]
            yield self._row_to_qapair(row)


class MultiHopRAGHF(SyftrQADataset):
    xname: T.Literal["multihoprag_hf"] = "multihoprag_hf"  # type: ignore

    description: str = (
        "This resource contains a corpus of news dataset. "
        "The news data source comprises various English-language "
        "websites covering a range of news categories: "
        "entertainment, business, sports, technology, "
        "health and science. It contains news articles "
        "on these topics published between September to December 2023. "
        "Every news article is paired with metadata, "
        "including title, publish date, author, category, URL and news source."
    )

    def _get_partition_range(self, partition: str):
        partition_ranges = {
            "sample": range(0, 20),
            "train": range(0, 1000),
            "test": range(1000, 2000),
            "holdout": range(2000, 2556),
        }
        if partition in partition_ranges:
            return partition_ranges[partition]

    def _load_grounding_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "yixuantt/MultiHopRAG",
                "corpus",
                cache_dir=cfg.paths.huggingface_cache,
            )
        # all of the corpus is in a split named "train"
        return dataset["train"]

    def _load_qa_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "yixuantt/MultiHopRAG",
                "MultiHopRAG",
                cache_dir=cfg.paths.huggingface_cache,
            )

        # all qapairs are in a split named "train"
        return dataset["train"]

    @overrides
    def iter_grounding_data(self, partition="notused") -> T.Iterator[Document]:
        # There is no partition. The grounding dataset is the same
        # across all partitions of the qa pairs.
        grounding_dataset = self._load_grounding_dataset()
        for row in grounding_dataset:
            yield Document(
                text=row["body"],
                metadata={
                    "title": row["title"],
                    "author": row["author"],
                    "category": row["category"],
                    "published_at": row["published_at"],
                    "url": row["url"],
                    "source": row["source"],
                },
            )

    def _row_to_qapair(self, row):
        """Dataset-specific conversion of row to QAPair struct.

        Invoked by iter_examples.

        Default implementation assumes row is already in QAPair format.
        """
        return QAPair(
            question=row["query"],
            answer=row["answer"],
            id=hashlib.md5(row["query"].encode()).hexdigest(),
            context={},
            supporting_facts=row["evidence_list"],
            difficulty="",
            qtype=str(row["question_type"]),
            gold_evidence=[item["fact"] for item in row["evidence_list"]],
        )

    @overrides
    def iter_examples(self, partition="test") -> T.Iterator[QAPair]:
        assert partition in self.storage_partitions
        partition = self._get_storage_partition(partition)
        qa_examples = self._load_qa_dataset()
        partition_range = self._get_partition_range(partition)

        for i in partition_range:
            row = qa_examples[i]
            yield self._row_to_qapair(row)


class BrightHF(SyftrQADataset):
    xname: T.Literal["bright_hf"] = "bright_hf"  # type: ignore

    # NOTE: pony, leetcode, aops, theoremqa_theorems, theoremqa_questions
    # subsets do not have ground truth answers
    subset: T.Literal[
        "earth_science",
        "biology",
        "economics",
        "psychology",
        "robotics",
        "stackoverflow",
        "sustainable_living",
        "pony",
        "leetcode",
        "aops",
        "theoremqa_theorems",
        "theoremqa_questions",
    ] = "biology"

    _descriptions = {
        "earth_science": (
            "This resource contains everything about earth science, including geology, "
            "oceanography, meteorology, and more. "
            "It contains historical data, live data, just about everything about earth science, for instance, "
            "information about geological formations, ocean currents, or weather patterns, "
            "key events or phenomena in earth science history, "
            "information about climate change or environmental issues, and "
            "information about earth science research methods or techniques."
        ),
        "biology": (
            "This resource contains everything about biology, including genetics, "
            "ecology, evolution, and more. "
            "It contains historical data, live data, just about everything about biology, for instance, "
            "information about cellular structures, ecosystems, or evolutionary processes, "
            "key events or discoveries in biology history, "
            "information about genetic engineering or biotechnology, and "
            "information about biological research methods or techniques."
        ),
        "economics": (
            "This resource contains everything about economics, including microeconomics, "
            "macroeconomics, international trade, and more. "
            "It contains historical data, live data, just about everything about economics, for instance, "
            "information about supply and demand, market structures, or economic indicators, "
            "key events or theories in economics history, "
            "information about economic policies or regulations, and "
            "information about economic research methods or techniques."
        ),
        "psychology": (
            "This resource contains everything about psychology, including cognitive, "
            "developmental, social, and clinical psychology. "
            "It contains historical data, live data, just about everything about psychology, for instance, "
            "information about brain structures, psychological disorders, or therapeutic techniques, "
            "key events or theories in psychology history, "
            "information about psychological research methods or techniques, and "
            "information about psychological assessments or tests."
        ),
        "robotics": (
            "This resource contains everything about robotics, including robot design, "
            "control systems, and applications. "
            "It contains historical data, live data, just about everything about robotics, for instance, "
            "information about robot components, sensors, or actuators, "
            "key events or advancements in robotics history, "
            "information about robotic applications in various fields, and "
            "information about robotics research methods or techniques."
        ),
        "stackoverflow": (
            "This resource contains everything about StackOverflow, including programming, "
            "software development, and more. "
            "It contains historical data, live data, just about everything about StackOverflow, for instance, "
            "information about programming languages, software frameworks, or development methodologies, "
            "key events or trends in software development history, "
            "information about software development best practices or tools, and "
            "information about software development research methods or techniques."
        ),
        "sustainable_living": (
            "This resource contains everything about sustainable living, including "
            "environmental conservation, renewable energy, and more. "
            "It contains historical data, live data, just about everything about sustainable living, for instance, "
            "information about sustainable practices, technologies, or policies, "
            "key events or movements in sustainability history, "
            "information about environmental issues or challenges, and "
            "information about sustainability research methods or techniques."
        ),
        "pony": (
            "This resource contains everything about the Pony programming language, including its syntax, "
            "capabilities, compiler architecture, and more. "
            "It contains documentation, examples, just about everything about Pony, for instance, "
            "information about Pony's actor-based concurrency model, capabilities-secure type system, "
            "key concepts like reference capabilities and behavior guarantees, "
            "information about pattern matching, generics, and other language features, and "
            "information about Pony's memory management, garbage collection, and runtime performance."
        ),
        "leetcode": (
            "This resource contains everything about LeetCode, including its problems, "
            "solutions, and more. "
            "It contains documentation, examples, just about everything about LeetCode, for instance, "
            "information about data structures, algorithms, or coding challenges, "
            "key concepts like time complexity and space complexity, "
            "information about problem-solving techniques and strategies, and "
            "information about LeetCode's online judge system and coding environment."
        ),
        "aops": (
            "This resource contains everything about Art of Problem Solving (AoPS), including its curriculum, "
            "problems, and more. "
            "It contains documentation, examples, just about everything about AoPS, for instance, "
            "information about problem-solving techniques, mathematical concepts, or competition preparation, "
            "key events or competitions in mathematics history, "
            "information about AoPS's online community and resources, and "
            "information about AoPS's books and courses."
        ),
        "theoremqa_theorems": (
            "This resource contains everything about TheoremQA, including its theorems, "
            "proofs, and more. "
            "It contains documentation, examples, just about everything about TheoremQA, for instance, "
            "information about mathematical theorems, proofs, or concepts, "
            "key events or advancements in theorem proving history, "
            "information about theorem proving techniques and strategies, and "
            "information about TheoremQA's online community and resources."
        ),
        "theoremqa_questions": (
            "This resource contains everything about TheoremQA, including its questions, "
            "answers, and more. "
            "It contains documentation, examples, just about everything about TheoremQA, for instance, "
            "information about mathematical questions, answers, or concepts, "
            "key events or advancements in theorem proving history, "
            "information about theorem proving techniques and strategies, and "
            "information about TheoremQA's online community and resources."
        ),
    }

    @property
    def description(self) -> str:
        return self._descriptions[self.subset]

    def _get_partition_range(self, partition: str):
        subset_partition_ranges = {
            "earth_science": {
                "sample": range(0, 10),
                "train": range(0, 30),
                "test": range(30, 90),
                "holdout": range(90, 116),
            },
            "biology": {
                "sample": range(0, 10),
                "train": range(0, 30),
                "test": range(30, 90),
                "holdout": range(90, 103),
            },
            "economics": {
                "sample": range(0, 10),
                "train": range(0, 30),
                "test": range(30, 90),
                "holdout": range(90, 103),
            },
            "psychology": {
                "sample": range(0, 10),
                "train": range(0, 30),
                "test": range(30, 90),
                "holdout": range(90, 101),
            },
            "robotics": {
                "sample": range(0, 10),
                "train": range(0, 30),
                "test": range(30, 90),
                "holdout": range(90, 101),
            },
            "stackoverflow": {
                "sample": range(0, 10),
                "train": range(0, 30),
                "test": range(30, 90),
                "holdout": range(90, 117),
            },
            "sustainable_living": {
                "sample": range(0, 10),
                "train": range(0, 30),
                "test": range(30, 90),
                "holdout": range(90, 108),
            },
            "pony": {
                "sample": range(0, 10),
                "train": range(0, 30),
                "test": range(30, 90),
                "holdout": range(90, 112),
            },
            "leetcode": {
                "sample": range(0, 10),
                "train": range(0, 30),
                "test": range(30, 120),
                "holdout": range(120, 142),
            },
            "aops": {
                "sample": range(0, 10),
                "train": range(0, 30),
                "test": range(30, 90),
                "holdout": range(90, 111),
            },
            "theoremqa_theorems": {
                "sample": range(0, 10),
                "train": range(0, 20),
                "test": range(20, 50),
                "holdout": range(50, 76),
            },
            "theoremqa_questions": {
                "sample": range(0, 10),
                "train": range(0, 40),
                "test": range(40, 160),
                "holdout": range(160, 194),
            },
        }

        partition_ranges = subset_partition_ranges[self.subset]
        if partition in partition_ranges:
            return partition_ranges[partition]

    def _load_grounding_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "xlangai/BRIGHT",
                "documents",
                cache_dir=cfg.paths.huggingface_cache,
            )
        # all of the corpus is in a split named "train"
        return dataset[self.subset]

    def _load_qa_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "xlangai/BRIGHT",
                "examples",
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset[self.subset]

    @overrides
    def iter_grounding_data(self, partition="notused") -> T.Iterator[Document]:
        # There is no partition. The grounding dataset is the same
        # across all partitions of the qa pairs.
        grounding_dataset = self._load_grounding_dataset()
        for row in grounding_dataset:
            yield Document(
                text=row["content"],
                metadata={
                    "id": row["id"],
                },
            )

    def _row_to_qapair(self, row):
        """Dataset-specific conversion of row to QAPair struct.

        Invoked by iter_examples.

        Default implementation assumes row is already in QAPair format.
        """
        return QAPair(
            question=row["query"],
            answer=row["gold_answer"],
            id=row["id"],
            context={},
            supporting_facts=[],
            difficulty="",
            qtype="",
        )

    @overrides
    def iter_examples(self, partition="test") -> T.Iterator[QAPair]:
        assert partition in self.storage_partitions
        partition = self._get_storage_partition(partition)
        qa_examples = self._load_qa_dataset()
        partition_range = self._get_partition_range(partition)

        for i in partition_range:
            row = qa_examples[i]
            yield self._row_to_qapair(row)


class PhantomWikiV001HF(SyftrQADataset):
    xname: T.Literal["phantomwikiv001_hf"] = "phantomwikiv001_hf"  # type: ignore

    # subset choices are:
    # depth_20_size_100_seed_595    depth_20_size_2000_seed_595   depth_20_size_400_seed_595
    # depth_20_size_1000_seed_595   depth_20_size_250_seed_595    depth_20_size_4000_seed_595
    # depth_20_size_10000_seed_595  depth_20_size_2500_seed_595   depth_20_size_450_seed_595
    # depth_20_size_100000_seed_595 depth_20_size_300_seed_595    depth_20_size_4500_seed_595
    # depth_20_size_150_seed_595    depth_20_size_3000_seed_595   depth_20_size_50_seed_595
    # depth_20_size_1500_seed_595   depth_20_size_350_seed_595    depth_20_size_500_seed_595
    # depth_20_size_200_seed_595    depth_20_size_3500_seed_595   depth_20_size_5000_seed_595

    subset: str = "depth_20_size_1000_seed_595"

    description: str = (
        "This dataset contains data from PhantomWiki, which "
        "is a framework for generating unique, factually"
        "and consistent document corpora with diverse question-answer pairs."
        "Unlike prior work, PhantomWiki is neither a fixed dataset, nor is it"
        "based on any existing data. Instead, a new PhantomWiki instance is "
        "generated on demand for each evaluation. PhantomWiki generates a "
        "fictional universe of characters along with a set of facts. "
        "We reflect these facts in a large-scale corpus, mimicking the "
        "style of fan-wiki websites. Then we generate question-answer pairs "
        "with tunable difficulties, encapsulating the types of multi-hop "
        "questions commonly considered in the question-answering (QA) literature."
    )

    def _get_partition_range(self, partition: str):
        partition_ranges = {
            "sample": range(0, 10),
            "train": range(0, 100),
            "test": range(100, 400),
            "holdout": range(400, 500),
        }
        if partition in partition_ranges:
            return partition_ranges[partition]

    def _load_grounding_dataset(self) -> datasets.Dataset:
        subset_name = self.subset + "_groundingdata"
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/phantomwiki-001",
                subset_name,
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset["train"]

    def _load_qa_dataset(self) -> datasets.Dataset:
        subset_name = self.subset + "_qapairs"
        # Load the dataset with the specified subset name
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/phantomwiki-001",
                subset_name,
                cache_dir=cfg.paths.huggingface_cache,
            )

        return dataset["train"]

    @overrides
    def iter_grounding_data(self, partition="notused") -> T.Iterator[Document]:
        # There is no partition. The grounding dataset is the same
        # across all partitions of the qa pairs.
        grounding_dataset = self._load_grounding_dataset()
        for row in grounding_dataset:
            yield Document(
                text=row["article"],
                metadata={"title": row["title"]},
            )

    def _row_to_qapair(self, row):
        """Dataset-specific conversion of row to QAPair struct.

        Invoked by iter_examples.

        Default implementation assumes row is already in QAPair format.
        """
        return QAPair(
            question=row["question"],
            answer=" ".join(row["answer"]),
            id=str(row["id"]),
            context={},
            supporting_facts=[],
            difficulty=str(row["difficulty"]),
            qtype=str(row["type"]),
        )

    @overrides
    def iter_examples(self, partition="test") -> T.Iterator[QAPair]:
        assert partition in self.storage_partitions
        partition = self._get_storage_partition(partition)
        qa_examples = self._load_qa_dataset()
        partition_range = self._get_partition_range(partition)

        for i in partition_range:
            row = qa_examples[i]
            yield self._row_to_qapair(row)


class JudgeEvalHF(SyftrQADataset):
    xname: T.Literal["judge_eval"] = "judge_eval"  # type: ignore

    description: str = (
        "Contains human labeled Question-Answer-Response triplets "
        "for evaluation of LLM-a-judge flows."
    )

    def _get_partition_range(self, partition: str):
        partition_ranges = {
            "sample": range(0, 20),
            "train": range(0, 300),
            "test": range(300, 700),
            "holdout": range(700, 807),
        }
        if partition in partition_ranges:
            return partition_ranges[partition]

    def _load_grounding_dataset(self) -> datasets.Dataset:
        raise NotImplementedError("judge_eval does not have a grounding dataset")

    def _load_qa_dataset(self) -> datasets.Dataset:
        with distributed_lock(
            self.name, timeout_s=self.load_examples_timeout_s, host_only=True
        ):
            dataset = datasets.load_dataset(
                "DataRobot-Research/judge-eval",
                cache_dir=cfg.paths.huggingface_cache,
            )

        # all qapairs are in a split named "train"
        return dataset["train"]

    @overrides
    def iter_grounding_data(self, partition="notused") -> T.Iterator[Document]:
        raise NotImplementedError("judge_eval does not have a grounding dataset")

    def _row_to_qartriplet(self, row):
        """Dataset-specific conversion of row to QARTriplet."""
        return QARTriplet(
            question=row["question"],
            answer=row["answer"],
            id=str(row["index"]),
            context={},
            supporting_facts=[],
            difficulty="",
            qtype="",
            gold_evidence=[],
            response=row["response"],
            label=bool(int(float(row["Human Passing"]))),
        )

    @overrides
    def iter_examples(self, partition="test") -> T.Iterator[QAPair]:
        assert partition in self.storage_partitions
        partition = self._get_storage_partition(partition)
        qa_examples = self._load_qa_dataset()
        partition_range = self._get_partition_range(partition)

        for i in partition_range:
            row = qa_examples[i]
            yield self._row_to_qartriplet(row)


class CustomDataset(SyftrQADataset):
    """
    Custom dataset class that loads data from local files.

    This class allows you to evaluate a RAG pipeline using local documents and QA pairs
    stored on your machine.

    The qa_csv and grounding_data_path fields must be absolute paths so that they can be
    resolved by both Ray workers and local scripts.

    The CustomDataset class is not intended for use in distributed Ray environments. To
    use this dataset in Ray, you are responsible for ensuring the dataset exists at the
    specified path on all worker nodes.

    In your study YAML file, specify this dataset as follows:

        dataset:
          xname: custom_dataset                             # Must use this exact string
          qa_csv_path: "/path/to/dataset.csv"               # Must be an absolute path
          grounding_data_dir: "/path/to/grounding/data/"    # Must be an absolute path
    """

    xname: T.Literal["custom_dataset"] = "custom_dataset"  # type: ignore

    qa_csv_path: Path = Path("/tmp/custom_qa_data.csv")
    grounding_data_dir: Path = Path("/tmp/grounding_docs")

    description: str = (
        "Custom local dataset for testing purposes."
        "This dataset loads QA pairs from a local CSV file and grounding data from a local directory."
    )

    sample_size: int = 10
    training_size: int = 100
    testing_size: int = 100
    holdout_size: int = 0

    def _get_partition_range(self, partition: str):
        partition_ranges = {
            "sample": range(0, self.sample_size),
            "train": range(0, self.training_size),
            "test": range(self.training_size, self.training_size + self.testing_size),
            "holdout": range(
                self.training_size + self.testing_size,
                self.training_size + self.testing_size + self.holdout_size,
            ),
        }
        if partition in partition_ranges:
            return partition_ranges[partition]
        raise ValueError(f"Unknown partition: {partition}")

    @field_validator("qa_csv_path", "grounding_data_dir", mode="after")
    @classmethod
    def check_paths_are_absolute(cls, v: Path) -> Path:
        if not v.is_absolute():
            raise ValueError(
                f"The provided dataset path `{v}` must be an absolute path"
            )
        return v

    @cached_property
    def data_root(self) -> Path:
        return Path(get_ray_temp_dir()) / "data" / "local_data"

    def _get_qa_abspath(self) -> Path:
        """
        Ray worker-visible QA CSV path.
        """
        return self.qa_csv_path.absolute()

    def _get_grounding_dir_abspath(self) -> Path:
        """
        Ray worker-visible grounding directory path.
        """
        return self.grounding_data_dir.absolute()

    def _load_qa_dataset(self, partition: str = "test") -> datasets.Dataset:
        """
        This method reads a CSV file of QA pairs and returns data of the requested partition.
        """

        df: pd.DataFrame = pd.read_csv(self._get_qa_abspath())
        partition_range = self._get_partition_range(partition)
        df_partition = df.iloc[list(partition_range)]
        return datasets.Dataset.from_pandas(df_partition)

    def _load_grounding_dataset(self) -> datasets.DatasetDict:
        reader = SimpleDirectoryReader(input_dir=self._get_grounding_dir_abspath())
        docs = reader.load_data()
        dataset = datasets.Dataset.from_pandas(
            pd.DataFrame(
                [
                    {
                        "markdown": doc.text,
                        "filename": doc.metadata.get("file_name", "unknown"),
                    }
                    for doc in docs
                ]
            )
        )
        return datasets.DatasetDict({"test": dataset})

    def _row_to_qapair(self, row) -> QAPair:
        """Dataset-specific conversion of row to QAPair struct."""
        return QAPair(
            question=row["question"],
            answer=row["answer"],
            id=str(row["id"]),
            context={},
            supporting_facts=[],
            difficulty="default",
            qtype="default",
        )

    @overrides
    def iter_examples(self, partition="test") -> T.Iterator[QAPair]:
        assert partition in self.storage_partitions
        partition = self._get_storage_partition(partition)
        qa_dataset = self._load_qa_dataset(partition)
        for row in qa_dataset:
            yield self._row_to_qapair(row)

    @overrides
    def iter_grounding_data(self, partition="test") -> T.Iterator[Document]:
        grounding_dataset = self._load_grounding_dataset()
        for row in grounding_dataset["test"]:
            yield Document(
                text=row["markdown"],
                metadata={"file_name": row["filename"]},
            )



import os
from pathlib import Path

from llama_index.core import VectorStoreIndex
from llama_index.core.ingestion import IngestionPipeline
from llama_index.core.node_parser import (
    SentenceSplitter,
)

from syftr.configuration import cfg
from syftr.hf_endpoint_models import get_hf_endpoint_embed_model


def test_hf_endpoint_vdb_creation(tiny_dataset):
    # Dummy line to load credentials
    _ = Path(cfg.paths.test_studies_dir / "test-no-rag-seeding.yaml")

    chunk_size = 1024
    chunk_overlap = 128
    model_name = "thomaskim1130/stella_en_400M_v5-FinanceRAG-v2"

    hf_embedding_model = get_hf_endpoint_embed_model(model_name)

    splitter = SentenceSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)
    transforms = [splitter]

    pipeline = IngestionPipeline(transformations=transforms)
    docs = list(tiny_dataset.iter_grounding_data())

    # Check if docs is not empty
    assert docs, "No documents loaded from dataset"

    nodes = pipeline.run(
        documents=docs,
        show_progress=True,
        max_workers=os.cpu_count(),
    )

    # Take only first 128 nodes
    nodes = nodes[:128]

    # Check if nodes is not empty
    assert nodes, "No nodes created from documents"

    # remove nodes with empty text
    nodes = [node for node in nodes if node.text]

    # Check if nodes have text
    assert all(node.text for node in nodes), "Some nodes have empty text"

    # create VectorStoreIndex using this embedding model
    vdb = VectorStoreIndex(
        nodes=nodes,
        embed_model=hf_embedding_model,
        insert_batch_size=2048,
        show_progress=True,
    )

    # Check if vdb is not None
    assert vdb is not None, "VectorStoreIndex creation failed"

    # Check the type of vdb
    assert isinstance(vdb, VectorStoreIndex), (
        "vdb is not an instance of VectorStoreIndex"
    )



from pathlib import Path

import pandas as pd

from syftr.baselines import set_baselines
from syftr.configuration import cfg
from syftr.optimization import StudyRunner
from syftr.optuna_helper import get_completed_trials
from syftr.ray.utils import ray_init
from syftr.studies import StudyConfig
from syftr.tuner.qa_tuner import objective, run_flow

PARAM_LIST = ["llm_name", "template_name", "rag_mode"]


def test_eval_pruning() -> None:
    study_config_file = Path(cfg.paths.test_studies_dir / "test-eval-pruning.yaml")
    study_config = StudyConfig.from_file(study_config_file)
    study_config.optimization.use_hf_embedding_models = False

    cfg.ray.local = True

    ray_init()

    set_baselines(study_config)

    optimization = StudyRunner(
        objective=objective,
        study_config=study_config,
        seeder=run_flow,
    )
    study = optimization.run()

    df_trials: pd.DataFrame = get_completed_trials(study)

    assert df_trials.empty, "All trials should have been pruned"



import json
import re
import typing as T

from llama_index.core.evaluation.base import BaseEvaluator
from llama_index.core.evaluation.correctness import CorrectnessEvaluator
from llama_index.core.prompts import (
    ChatMessage,
    ChatPromptTemplate,
    MessageRole,
)

from syftr.llm import get_llm
from syftr.logger import logger
from syftr.studies import AgentStudyConfig, StudyConfig


def json_parser_function(response: str) -> T.Tuple[T.Optional[float], T.Optional[str]]:
    """
    Parse the response from the evaluator to extract score and reasoning.
    In case the response is not valid, it returns None for both score and reasoning.
    The last JSON-like substring of the response is expected to be non-nested and
    should contain a "score" and "reasoning" key.
    The score can be a float or a string that can be converted to a float.
    The reasoning is expected to be a string.
    """

    # Check if the response ends with a nested JSON
    if re.search(r"\{[^{}]*\{[^{}]*\}[^{}]*\}[^{}]*$", response):
        logger.error("Nested JSON found at the end of evaluator response: %s", response)
        return None, None

    # Find all JSON-like objects in the response
    json_pattern = r"\{[^{}]*\}"
    matches = re.findall(json_pattern, response)

    if matches:
        last_json_str = matches[-1]
        try:
            response_dict = json.loads(last_json_str)
        except json.JSONDecodeError:
            logger.error("Invalid JSON response from evaluator: %s", response)
            return None, None
    else:
        logger.error("No JSON found in evaluator response: %s", response)
        return None, None

    score = response_dict.get("score")
    if score is not None:
        try:
            score = float(score)
        except ValueError:
            logger.error("Invalid score in evaluator response: %s", response)
            return None, None
    else:
        logger.error("Score missing in evaluator response: %s", response)
        return None, None
    reasoning = response_dict.get("reasoning")
    if not reasoning or not isinstance(reasoning, str):
        logger.error("Invalid reasoning in evaluator response: %s", response)
        return None, None
    return score, reasoning


class CorrectnessEvaluatorFactory:
    """Factory class to create LLM judges of type CorrectnessEvaluator based on the study configuration."""

    def __init__(
        self,
        study_config: T.Union[StudyConfig, AgentStudyConfig],
    ):
        assert isinstance(study_config, StudyConfig), (
            "AgentStudyConfig needs to provide dataset information."
        )

        self.llm_names = study_config.evaluation.llm_names
        self.eval_type = study_config.evaluation.eval_type
        self.eval_system_template = study_config.evaluation.eval_system_template
        self.eval_user_template = study_config.dataset.eval_user_template
        self.score_threshold = study_config.evaluation.score_threshold

        assert self.eval_type == "correctness", (
            f"Unsupported evaluation type: {self.eval_type}. "
            "Currently only 'correctness' is supported."
        )

    def _get_correctness_evaluators(self) -> T.List[BaseEvaluator]:
        eval_llms = [get_llm(name) for name in self.llm_names]
        eval_template = ChatPromptTemplate(
            message_templates=[
                ChatMessage(role=MessageRole.SYSTEM, content=self.eval_system_template),
                ChatMessage(role=MessageRole.USER, content=self.eval_user_template),
            ]
        )
        return [
            CorrectnessEvaluator(
                llm=llm,
                eval_template=eval_template,
                score_threshold=self.score_threshold,
                parser_function=json_parser_function,
            )
            for llm in eval_llms
        ]

    def get_evaluators(self) -> T.List[BaseEvaluator]:
        match self.eval_type:
            case "correctness":
                return self._get_correctness_evaluators()
            case _:
                raise ValueError(
                    f"Unsupported evaluation type: {self.eval_type}. "
                    "Currently only 'correctness' is supported."
                )



import json
import re
import typing as T

import numpy as np
import optuna
import pandas as pd
import ray
from kneed import KneeLocator
from numpy import ceil
from optuna.storages import BaseStorage
from paretoset import paretoset
from ray._private.worker import RemoteFunction1
from tenacity import retry, stop_after_attempt, wait_fixed
from tqdm import tqdm

from syftr.configuration import NDIGITS, cfg
from syftr.hierarchical_tpe import HierarchicalTPESampler
from syftr.logger import logger
from syftr.studies import StudyConfig


def get_flows_from_trials(df: pd.DataFrame) -> T.List[T.Dict[str, T.Any]]:
    return df["user_attrs_flow"].apply(json.loads).tolist()


def get_study_names(
    include_regex: T.Union[str, T.List[str]],
    exclude_regex: T.Optional[T.Union[str, T.List[str]]] = None,
) -> T.List[str]:
    exclude_regex = exclude_regex or []
    include_regex = [include_regex] if isinstance(include_regex, str) else include_regex
    exclude_regex = [exclude_regex] if isinstance(exclude_regex, str) else exclude_regex
    all_study_names = optuna.get_all_study_names(
        storage=cfg.database.get_optuna_storage()
    )
    assert all_study_names, "No studies found"
    return sorted(
        [
            s
            for s in all_study_names
            if any(re.match(regex, s) for regex in include_regex)
            and not any(re.match(regex, s) for regex in exclude_regex)
        ]
    )


def without_non_search_space_params(
    params: T.Dict[str, T.Any], study_config: StudyConfig
) -> T.Dict[str, T.Any]:
    result = {}
    for k, v in params.items():
        if k not in study_config.search_space.non_search_space_params:
            result[k] = v
    return result


def trial_exists(
    study_name: str,
    params: T.Dict[str, T.Any],
    storage: str | BaseStorage | None = None,
):
    storage = storage or cfg.database.get_optuna_storage()
    logger.debug("Loading '%s' from storage: %s", study_name, storage)
    study = optuna.load_study(study_name=study_name, storage=storage)
    for trial in study.get_trials():
        if params == trial.params:
            return True
    return False


def recreate_with_completed_trials(
    study_config: StudyConfig,
    storage: str | BaseStorage | None = None,
):
    study_name = study_config.name
    storage = storage or cfg.database.get_optuna_storage()
    try:
        study: optuna.Study = optuna.load_study(study_name=study_name, storage=storage)
    except KeyError:
        logger.warning(
            "Cannot recreate study '%s' because it does not exist", study_name
        )
        return
    completed_trials = [
        optuna.trial.create_trial(
            state=optuna.trial.TrialState.COMPLETE,
            values=t.values,
            params=t.params,
            distributions=t.distributions,
            user_attrs=t.user_attrs,
            system_attrs=t.system_attrs,
            intermediate_values=t.intermediate_values,
        )
        for t in study.trials
        if t.state == optuna.trial.TrialState.COMPLETE
    ]
    logger.info("Keeping %i completed trials", len(completed_trials))
    logger.info("Deleting study '%s'", study_name)
    optuna.delete_study(study_name=study_name, storage=storage)
    logger.info("Recreating study '%s'", study_name)
    sampler = get_sampler(study_config)
    new_study = optuna.create_study(
        storage=storage,
        sampler=sampler,
        directions=["maximize", "minimize"],
        study_name=study_name,
        load_if_exists=False,
    )
    new_study.add_trials(completed_trials)
    logger.info("Recreated study '%s' successfully", study_name)


@retry(stop=stop_after_attempt(6), wait=wait_fixed(10))
def _get_completed_trials(
    study: optuna.Study,
    success_rate: float | None = None,
    ndigits: int = NDIGITS,
):
    df: pd.DataFrame = study.trials_dataframe()
    if df.empty:
        return df
    df = df[df["state"] == "COMPLETE"]
    if "user_attrs_metric_failed" in df:
        df = df[df["user_attrs_metric_failed"] == False]
    df = df.dropna(subset=["values_0", "values_1"])
    if success_rate:
        try:
            success = df["user_attrs_metric_num_success"]
            n_total = df["user_attrs_metric_num_success"].max()
            df = df[(success / n_total) >= success_rate]
        except KeyError:
            logger.warning(f"Cannot filter {study.study_name} by success rate yet")
    df = df.reset_index(drop=True)
    for col in df.select_dtypes(include=[np.float64]).columns:
        if col.startswith("params_"):
            df[col] = df[col].round(ndigits)
    return df


def get_completed_trials(
    study: T.Union[optuna.Study, str, T.List[optuna.Study], T.List[str]],
    storage: str | BaseStorage | None = None,
    success_rate: float | None = None,
    drop_inf_latency: bool | None = True,
) -> pd.DataFrame:
    study_list = study if isinstance(study, list) else [study]
    storage = storage or cfg.database.get_optuna_storage()
    tmp: T.List[optuna.Study] = []
    for s in study_list:
        if isinstance(s, str):
            tmp.append(optuna.load_study(study_name=s, storage=storage))
        else:
            assert isinstance(s, optuna.Study), "Bad study provided"
            tmp.append(s)
    studies = tmp
    dfs: T.List[pd.DataFrame] = []
    for s in tqdm(
        studies, desc="Loading trials from studies", disable=len(studies) <= 1
    ):
        df_tmp: pd.DataFrame = _get_completed_trials(
            study=s,  # type: ignore[arg-type]
            success_rate=success_rate,
        )
        if df_tmp is None:
            logger.warning("Trials dataframe should not be None")
            continue
        if not df_tmp.empty:
            df_tmp["study_name"] = s.study_name  # type: ignore[union-attr]
            dfs.append(df_tmp)
    if dfs:
        df_result = pd.concat(dfs, ignore_index=True)
        if drop_inf_latency:
            df_result = df_result[df_result["values_1"] < np.inf]
            df_result = df_result.reset_index(drop=True)
        logger.info(f"Loaded {len(df_result)} trials")
        return df_result
    return pd.DataFrame()


def get_completed_flows(
    study: T.Union[optuna.Study, str, T.List[optuna.Study], T.List[str]],
    storage: str | BaseStorage | None = None,
    success_rate: float | None = None,
    drop_inf_latency: bool | None = True,
) -> pd.DataFrame:
    df_trials: pd.DataFrame = get_completed_trials(
        study=study,
        storage=storage,
        success_rate=success_rate,
        drop_inf_latency=drop_inf_latency,
    )
    return get_flows_from_trials(df_trials)


def get_pareto_mask(
    study: optuna.Study | str | pd.DataFrame,
    success_rate: float | None = None,
    accuracy_col: str = "values_0",
    latency_col: str = "values_1",
) -> np.ndarray:
    if isinstance(study, pd.DataFrame):
        df_trials = study.copy()
        if success_rate:
            success = df_trials["user_attrs_metric_num_success"]
            total = df_trials["user_attrs_metric_num_total"]
            df_trials = df_trials[(success / total) >= success_rate]
    else:
        assert isinstance(study, (optuna.Study, str)), "Bad study provided"
        df_trials = get_completed_trials(study, success_rate=success_rate)
    return paretoset(df_trials[[accuracy_col, latency_col]], sense=["max", "min"])


def get_pareto_df(
    study_config: StudyConfig | str,
    success_rate: float | None = None,
    storage: str | BaseStorage | None = None,
) -> pd.DataFrame:
    """
    Get the pareto front of the study in form of a Pandas DataFrame.
    The DataFrame contains the trials that are on the pareto front.
    Args:
        study_config: The study config or the name of the study.
        success_rate: The success rate to filter the trials.
    """
    study_name = (
        study_config.name if isinstance(study_config, StudyConfig) else study_config
    )
    storage = storage or cfg.database.get_optuna_storage()
    study: optuna.Study = optuna.load_study(study_name=study_name, storage=storage)
    df_trials: pd.DataFrame = get_completed_trials(study, success_rate=success_rate)
    pareto_mask = get_pareto_mask(df_trials)
    return df_trials[pareto_mask]


def get_pareto_flows(
    study_config: StudyConfig | str, success_rate: float | None = None
) -> T.List[T.Dict[str, T.Any]]:
    df_pareto = get_pareto_df(study_config, success_rate=success_rate)
    return get_flows_from_trials(df_pareto)


def get_knee_df(
    study_config: StudyConfig, success_rate: float | None = None
) -> pd.DataFrame:
    df_pareto: pd.DataFrame = get_pareto_df(study_config, success_rate=success_rate)
    if df_pareto.empty:
        return df_pareto
    knee = KneeLocator(
        df_pareto["values_1"],
        df_pareto["values_0"],
        curve="concave",
        direction="increasing",
    )
    return knee.knee


def get_knee_flow(
    study_config: StudyConfig, success_rate: float | None = None
) -> T.Dict[str, T.Any]:
    df_knee = get_knee_df(study_config, success_rate=success_rate)
    if df_knee.empty:
        return {}
    return get_flows_from_trials(df_knee)[0]


def get_failed_trials(study_names: str | T.List[str]) -> pd.DataFrame:
    study_names = study_names if isinstance(study_names, list) else [study_names]
    storage = cfg.database.get_optuna_storage()
    dfs: T.List[pd.DataFrame] = []
    for study_name in study_names:
        study: optuna.Study = optuna.load_study(study_name=study_name, storage=storage)
        df: pd.DataFrame = study.trials_dataframe()
        if df.empty:
            continue
        if "user_attrs_metric_failed" not in df.columns:
            continue
        df = df[df["user_attrs_metric_failed"] == True]
        df["study_name"] = study.study_name
        dfs.append(df)
    if not dfs:
        return pd.DataFrame()
    return pd.concat(dfs, ignore_index=True)


def run_flows(
    flows: T.List[T.Dict[str, T.Any]],
    study_config: StudyConfig,
    runner: RemoteFunction1[None, dict[str, T.Any], StudyConfig],
):
    n_flows = len(flows)
    n_parallel = study_config.optimization.max_concurrent_trials
    n_batches = int(ceil(n_flows / n_parallel))

    num_cpus = study_config.optimization.cpus_per_trial
    num_gpus = study_config.optimization.gpus_per_trial
    logger.info(
        "Running %i flows (max %i in parallel) using study config: %s",
        n_flows,
        n_parallel,
        study_config.name,
    )
    for i in range(n_batches):
        batch_args = flows[i * n_parallel : (i + 1) * n_parallel]
        tasks = [
            runner.options(num_cpus=num_cpus, num_gpus=num_gpus).remote(b, study_config)
            for b in batch_args
        ]
        ray.get(tasks)


def seed_study(seeder, study_config: StudyConfig):
    flows = study_config.optimization.baselines
    if study_config.optimization.shuffle_baselines:
        logger.info("Shuffling baselines")
        flows = sorted(flows, key=lambda _: np.random.random())
    run_flows(flows, study_config, seeder)


def _set_metric(trial: optuna.trial.BaseTrial, metric_name: str, score: T.Any):
    trial.set_user_attr("metric_" + metric_name, score)


def set_metrics(trial: optuna.trial.BaseTrial, metrics: T.Dict[str, float] | None):
    assert metrics, "No metrics provided"
    for metric_name, score in metrics.items():
        _set_metric(trial, metric_name, score)


def recreate_locally(study: optuna.Study):
    logger.info("Creating in-memory study from existing study")
    new_study = optuna.create_study(
        study_name=study.study_name,
        directions=study.directions,
    )
    for trial in tqdm(study.trials, desc="Copying trials"):
        new_study.add_trial(trial)
    return new_study


def get_sampler(study_config: StudyConfig) -> optuna.samplers.BaseSampler:
    if study_config.optimization.sampler == "tpe":
        return optuna.samplers.TPESampler(
            n_startup_trials=study_config.optimization.num_random_trials,
            constant_liar=True,
            multivariate=True,
        )
    elif study_config.optimization.sampler == "hierarchical":
        return HierarchicalTPESampler(
            constant_liar=True,
            n_startup_trials=study_config.optimization.num_random_trials,
        )
    else:
        raise ValueError("Invalid sampler")



import argparse
import asyncio
import json
import time
import typing as T

from syftr.configuration import cfg
from syftr.experiments import iter_all_job_logs
from syftr.helpers import get_flows_from_trials
from syftr.logger import logger
from syftr.optimization import user_confirm_delete
from syftr.optuna_helper import (
    get_completed_trials,
    get_pareto_flows,
)
from syftr.ray.submit import get_client, start_study
from syftr.storage import (  # noqa
    BrightHF,
    CragTask3HF,
    DRDocsHF,
    FinanceBenchHF,
    HotPotQAHF,
    InfiniteBenchHF,
    MultiHopRAGHF,
    PhantomWikiv050,
    SyftrQADataset,
    SyntheticCragTask3HF,
    SyntheticFinanceBenchHF,
    SyntheticHotPotQAHF,
)
from syftr.studies import (  # noqa
    LOCAL_EMBEDDING_MODELS,
    Block,
    CritiqueRAGAgent,
    Evaluation,
    FewShotRetriever,
    Hyde,
    LATSRagAgent,
    LLMConfig,
    OptimizationConfig,
    QueryDecomposition,
    ReactRAGAgent,
    Reranker,
    Retriever,
    SearchSpace,
    Splitter,
    StudyConfig,
    SubQuestionRAGAgent,
    TimeoutConfig,
    TopK,
    TransferLearningConfig,
)
from syftr.studyconfig_helper import build_configs

# -------------------------------------------------------
PREFIX = "gpt-oss"  # this three parameters
BENCH_NUM = 1  # are used to name
# RUN_NAME = "thinking"
RUN_NAME = "cost"
# -------------------------------------------------------
NUM_TRIALS = 3000  # total number of optimization trials per submission
NUM_RANDOM_TRIALS = 100
MAX_CONCURRENT_TRIALS = 20
NUM_EVAL_SAMPLES = 50
REUSE_STUDY = True  # WARNING: if set to False, existing studies will be deleted!
RECREATE_STUDY = (
    True  # WARNING: do not use with simultaneous runs using the same study!
)
EVAL_MODE: T.Literal["single", "random", "consensus"] = "single"
DRY_RUN = False  #  a dry run will not submit jobs but create the study configs
EMBEDDING_MAX_TIME = 3600 * 8
MINUTES_BEFORE_NEXT_SUBMISSION = 1
OBJ2_NAME = "llm_cost_mean"  # "p80_time", "llm_cost_mean", "retriever_context_length"
# -------------------------------------------------------
# To seed with silver bullets, you first create the input file using silver_bullets.ipynb notebook
CUSTOM_BASELINES = None  # valid values are: "pareto", "all", "silver", "transfer", None  # valid values are: "pareto", "all", "silver", "tansfer", None
BASELINES_BATCH_SIZE = 100  # we require batching of baselines to avoid Ray OOM issues
BASELINES_START = 0  # you can restrict the number of baselines ...
BASELINES_END = 100  # ... to start with here to avoid OOM issues
BASELINE_STUDIES: T.List[str] = [
    # "seeding1--training--crag_hf-music--music",
    # "seeding1--training--financebench_hf",
    # "seeding1--training--hotpotqa_hf-train_hard--train_hard",
    # "seeding1--training--multihoprag_hf",
]

BLOCKS = [
    Block(
        name="global",
        num_trials=NUM_TRIALS,
        components=[
            "rag_retriever",
            "splitter",
            "additional_context",
            "few_shot_retriever",
            "hyde",
            "critique_rag_agent",
            "lats_rag_agent",
            "react_rag_agent",
            "rag_mode",
            "reranker",
            "response_synthesizer",
            "sub_question_rag",
            "template_name",
        ],
    ),
    # Block(
    #     name="rag_retriever",
    #     num_trials=200,
    #     components=["rag_retriever"],
    # ),
    # Block(
    #     name="main",
    #     num_trials=NUM_TRIALS - 200,
    #     components=[
    #         "splitter",
    #         "additional_context",
    #         "few_shot_retriever",
    #         "hyde",
    #         "critique_rag_agent",
    #         "lats_rag_agent",
    #         "react_rag_agent",
    #         "rag_mode",
    #         "reranker",
    #         "response_synthesizer",
    #         "sub_question_rag",
    #         "template_name",
    #     ],
    # ),
]

TRANSFER_LEARNING = None
BASELINES = []
if CUSTOM_BASELINES == "pareto":
    for study in BASELINE_STUDIES:
        for flow in get_pareto_flows(study, 0.9):
            if flow not in BASELINES:
                BASELINES.append(flow)
    logger.info(f"We have {len(BASELINES)} Pareto-baselines for seeding")
elif CUSTOM_BASELINES == "all":
    df_trials = get_completed_trials(study=BASELINE_STUDIES)
    df_trials = df_trials.sort_values(by="number")
    flows = get_flows_from_trials(df_trials)
    BASELINES.extend(flows)
    logger.info(f"We have {len(BASELINES)} baselines for seeding")
elif CUSTOM_BASELINES == "silver":
    BASELINES = json.load(open(cfg.paths.results_dir / "silver-bullets.json", "r"))
    logger.info(f"We have {len(BASELINES)} silver bullet baselines for seeding")
elif CUSTOM_BASELINES == "transfer":
    TRANSFER_LEARNING = TransferLearningConfig(
        studies=BASELINE_STUDIES,
        max_fronts=2,
        max_total=23,
        success_rate=0.9,
        embedding_model="BAAI/bge-large-en-v1.5",
    )
else:
    logger.info("No custom baselines provided")


LLMS: T.List[str] = [
    "qwen3-235b-a22b",
    "glm-4.5-air",
    "gpt-oss-120b-low",
    "gpt-oss-120b-medium",
    "gpt-oss-120b-high",
    "gpt-oss-20b-low",
    "gpt-oss-20b-medium",
    "gpt-oss-20b-high",
    "nemotron-super-49b",
    "qwen3-30b-a3b",
    "gemma3-27b-it",
    "phi-4-multimodal-instruct",
]


EMBEDDING_MODELS = [
    "BAAI/bge-small-en-v1.5",
    "thenlper/gte-large",
    "mixedbread-ai/mxbai-embed-large-v1",
    "WhereIsAI/UAE-Large-V1",
    "avsolatorio/GIST-large-Embedding-v0",
    "w601sxs/b1ade-embed",
    "Labib11/MUG-B-1.6",
    "sentence-transformers/all-MiniLM-L12-v2",
    "sentence-transformers/paraphrase-multilingual-mpnet-base-v2",
    "BAAI/bge-base-en-v1.5",
    "baconnier/Finance2_embedding_small_en-V1.5",
    "FinLang/finance-embeddings-investopedia",
    "BAAI/bge-large-en-v1.5",
    # "BAAI/bge-multilingual-gemma2",
    "Snowflake/snowflake-arctic-embed-l-v2.0",
    "TencentBAC/Conan-embedding-v1",
    "amentaphd/snowflake-artic-embed-l",
]

SEARCH_SPACE = SearchSpace(
    few_shot_enabled=[False, True],
    additional_context_enabled=[False, True],
    hyde_enabled=[False, True],
    reranker_enabled=[False, True],
    splitter=Splitter(
        methods=[
            "recursive",
            "sentence",
            "token",
        ],
        chunk_min_exp=8,
        chunk_max_exp=10,
        chunk_overlap_frac_min=0.0,
        chunk_overlap_frac_max=0.5,
        chunk_overlap_frac_step=0.05,
    ),
    rag_modes=[
        # "no_rag",
        "rag",
        "lats_rag_agent",
        "react_rag_agent",
        "critique_rag_agent",
        "sub_question_rag",
    ],
    template_names=[
        "default",
        "concise",
        "CoT",
        # "finance-expert",
    ],
    response_synthesizer_llm_config=LLMConfig(llm_names=LLMS),
    rag_retriever=Retriever(
        embedding_models=EMBEDDING_MODELS,
        methods=["dense", "sparse", "hybrid"],
        top_k=TopK(kmin=1, kmax=10, log=False),
        query_decomposition=QueryDecomposition(
            llm_config=LLMConfig(llm_names=LLMS),
            num_queries_min=2,
            num_queries_max=5,
            num_queries_step=1,
        ),
    ),
    react_rag_agent=ReactRAGAgent(
        subquestion_engine_llm_config=LLMConfig(llm_names=LLMS),
        subquestion_response_synthesizer_llm_config=LLMConfig(llm_names=LLMS),
    ),
    sub_question_rag=SubQuestionRAGAgent(
        subquestion_engine_llm_config=LLMConfig(llm_names=LLMS),
        subquestion_response_synthesizer_llm_config=LLMConfig(llm_names=LLMS),
    ),
    critique_rag_agent=CritiqueRAGAgent(
        subquestion_engine_llm_config=LLMConfig(llm_names=LLMS),
        subquestion_response_synthesizer_llm_config=LLMConfig(llm_names=LLMS),
        critique_agent_llm_config=LLMConfig(llm_names=LLMS),
        reflection_agent_llm_config=LLMConfig(llm_names=LLMS),
    ),
    lats_rag_agent=LATSRagAgent(),
    reranker=Reranker(llm_config=LLMConfig(llm_names=LLMS)),
    hyde=Hyde(llm_config=LLMConfig(llm_names=LLMS)),
    few_shot_retriever=FewShotRetriever(
        embedding_models=EMBEDDING_MODELS,
    ),
)

EVALUATION = Evaluation(
    mode=EVAL_MODE,
    raise_on_exception=False,
)
# EVALUATION.llm_names = LLMS
EVALUATION.llm_names = ["gpt-4o-mini"]

DATASETS: T.List[SyftrQADataset] = [
    # CragTask3HF(subset="music"),
    FinanceBenchHF(),
    HotPotQAHF(subset="train_hard"),
    MultiHopRAGHF(),
    # -----------------------------------------------
    # BrightHF(subset="biology"),
    # DRDocsHF(),
    # InfiniteBenchHF(),
    PhantomWikiv050(),
    # ###############################################
    # BrightHF(subset="earth_science"),
    # BrightHF(subset="economics"),
    # BrightHF(subset="pony"),
    # BrightHF(subset="psychology"),
    # BrightHF(subset="robotics"),
    # BrightHF(subset="stackoverflow"),
    # BrightHF(subset="sustainable_living"),
    # CragTask3HF(subset="sports"),
]
assert DATASETS, "No datasets found. Please check the dataset list."


def get_optimization_parameters():
    optimization_config = OptimizationConfig(
        method="expanding",
        blocks=BLOCKS,
        shuffle_blocks=False,
        num_trials=NUM_TRIALS,
        baselines=BASELINES,
        baselines_cycle_llms=True,
        shuffle_baselines=True,
        max_concurrent_trials=MAX_CONCURRENT_TRIALS,
        num_eval_samples=NUM_EVAL_SAMPLES,
        num_eval_batch=5,
        rate_limiter_max_coros=10,  # control the number of concurrent evals ...
        rate_limiter_period=60,  # ... per given time unit
        max_trial_cost=40.0,
        cpus_per_trial=1,
        seeder_timeout=None,  # None: wait until finished, 0: don't wait
        # -----------------------------------------------
        num_random_trials=NUM_RANDOM_TRIALS,
        # -----------------------------------------------
        use_individual_baselines=False,
        use_agent_baselines=False,
        use_variations_of_baselines=False,
        # -----------------------------------------------
        use_pareto_baselines=False,  # required for transfer learning
        # -----------------------------------------------
        use_pareto_pruner=True,
        use_cost_pruner=True,
        use_runtime_pruner=True,
        # -----------------------------------------------
        use_toy_baselines=False,
        # -----------------------------------------------
        sampler="tpe",
        objective_2_name=OBJ2_NAME,
    )
    if BASELINES:
        start = BASELINES_START or 0
        end = BASELINES_END or len(BASELINES)
        for i in range(start, end, BASELINES_BATCH_SIZE):
            optimization_config = optimization_config.model_copy()
            optimization_config.baselines = BASELINES[i : i + BASELINES_BATCH_SIZE]
            yield DATASETS, SEARCH_SPACE, optimization_config, EVALUATION
    else:
        yield DATASETS, SEARCH_SPACE, optimization_config, EVALUATION


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--remote",
        help="Use remote Ray cluster",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--number",
        type=int,
        default=BENCH_NUM,
        help="The benchmark number used to set up configurations",
    )
    args = parser.parse_args()
    cfg.ray.local = False if args.remote else cfg.ray.local

    job_ids = []
    client = get_client()
    for (
        datasets,
        search_space,
        optimization_config,
        evaluation,
    ) in get_optimization_parameters():
        configs, paths = build_configs(
            datasets=datasets,
            search_space=search_space,
            optimization_config=optimization_config,
            evaluation=evaluation,
            bench_num=args.number,
            reuse_study=REUSE_STUDY,
            recreate_study=RECREATE_STUDY,
            prefix=PREFIX,
            run_name=RUN_NAME,
            embedding_max_time=EMBEDDING_MAX_TIME,
            transfer_learning=TRANSFER_LEARNING,
        )

        if DRY_RUN:
            print("Not submitting jobs because DRY_RUN is set to True")
            continue

        delete_confirmed = user_confirm_delete(configs[0])

        # launch benchmarks
        assert delete_confirmed

        for i, (config, path) in enumerate(zip(configs, paths)):
            job_id = start_study(
                client, path, config, delete_confirmed=delete_confirmed
            )
            job_ids.append(job_id)
            logger.info("Started job %s", job_id)
            # This might help the checkpointing bug
            sleep_time = 60 * MINUTES_BEFORE_NEXT_SUBMISSION
            logger.info(
                f"Sleeping for {MINUTES_BEFORE_NEXT_SUBMISSION} minutes before the next submission"
            )
            time.sleep(int(sleep_time))

    # monitor benchmarks
    log_tailers = [client.tail_job_logs(job) for job in job_ids]

    asyncio.run(iter_all_job_logs(log_tailers))  # type: ignore


if __name__ == "__main__":
    main()



import socket
from multiprocessing import Pool

from syftr.huggingface_helper import get_embedding_model
from syftr.storage import HotPotQAHF, PartitionMap
from syftr.studies import StudyConfig

hostname = socket.gethostname()
study_config = StudyConfig(
    name="create-onnx",
    dataset=HotPotQAHF(
        partition_map=PartitionMap(test="sample"),
    ),
)
search_space = study_config.search_space


def process_model(name):
    model, is_onnx = get_embedding_model(name, study_config.timeouts, total_chunks=0)
    assert model, f"Cannot get a model for '{name}' on {hostname}"
    assert is_onnx in [True, False], "Bad is_onnx response"
    model.get_query_embedding("Welcome!")
    if is_onnx:
        print(f"Create ONNX version of '{name}' on {hostname}")
    else:
        print(f"ONNX version of '{name}' not available on {hostname}")


if __name__ == "__main__":
    with Pool() as pool:
        pool.map(process_model, search_space.rag_retriever.embedding_models)



from optuna.visualization import (
    plot_contour,
    plot_optimization_history,
    plot_parallel_coordinate,
    plot_param_importances,
    plot_pareto_front,
    plot_slice,
)

from syftr.configuration import cfg


def plot_accuracy_and_wc_time_w_optuna(study):
    plot_optimization_history(
        study,
        target=lambda t: t.values[0],
        target_name=cfg.plotting.target_accuracy_name,
    ).show()

    plot_optimization_history(
        study,
        target=lambda t: t.values[1],
        target_name=cfg.plotting.target_latency_name,
    ).show()

    plot_parallel_coordinate(
        study,
        target=lambda t: t.values[0],
        target_name=cfg.plotting.target_accuracy_name,
    ).show()

    plot_parallel_coordinate(
        study,
        target=lambda t: t.values[1],
        target_name=cfg.plotting.target_latency_name,
    ).show()

    plot_slice(
        study,
        target=lambda t: t.values[0],
        target_name=cfg.plotting.target_accuracy_name,
    ).show()

    plot_slice(
        study,
        target=lambda t: t.values[1],
        target_name=cfg.plotting.target_latency_name,
    ).show()

    plot_contour(
        study,
        target=lambda t: t.values[0],
        target_name=cfg.plotting.target_accuracy_name,
    ).show()

    plot_contour(
        study,
        target=lambda t: t.values[1],
        target_name=cfg.plotting.target_latency_name,
    ).show()

    plot_param_importances(
        study,
        target=lambda t: t.values[0],
        target_name=cfg.plotting.target_accuracy_name,
    )

    plot_param_importances(
        study,
        target=lambda t: t.values[1],
        target_name=cfg.plotting.target_latency_name,
    )

    plot_pareto_front(
        study=study,
        target_names=[
            cfg.plotting.target_latency_name,
            cfg.plotting.target_accuracy_name,
        ],
        targets=lambda t: (t.values[1], t.values[0]),
    ).show()



import typing as T

import numpy as np
from llama_index.core.evaluation.retrieval.metrics_base import (
    BaseRetrievalMetric,
    RetrievalMetricResult,
)
from rouge_score import rouge_scorer


class RougeLRecallRetrievalMetric(BaseRetrievalMetric):
    """Custom retrieval metric using RougeL precision between expected and retrieved texts.

    Given expected text strings and a set of retrieved texts which may contain some or all of the expected
    string(s), compute the ROUGE-L precision between the retrieved and expected texts.

    Precision is 1 when the entire expected text is contained in the retrieved text, 0.90 when 90% is contained, etc.

    For each expected text we find the retrieved text with the highest precision to estimate the recall score for that text.
    The recall is then the average recall score across the expected texts.

    e.g. if we have three expected texts and ten retrieved texts, and the best precisions for each are
        [0.95, 0.30, 0.85]
    then the recall is
        sum([0.95, 0.30, 0.85]) / 3.0

    The recall is 1 if all expected texts are retrieved, 0 if there is no overlap, etc.

    This is a useful metric when the following conditions hold:
    - The IDs of the expected texts are not known
    - The expected texts are shorter in length than the retrieved documents, but still around 1+ complete sentences
    - The expected texts may be cut off between two documents in the retrieval index.
    """

    metric_name = "RougeLRecall"

    def compute(
        self,
        query: T.Optional[str] = None,
        expected_ids: T.Optional[T.List[str]] = None,
        retrieved_ids: T.Optional[T.List[str]] = None,
        expected_texts: T.Optional[T.List[str]] = None,
        retrieved_texts: T.Optional[T.List[str]] = None,
        **kwargs: T.Any,
    ) -> RetrievalMetricResult:
        """Compute metric.

        Args:
            query (Optional[str]): Query string
            expected_ids (Optional[List[str]]): Expected ids
            retrieved_ids (Optional[List[str]]): Retrieved ids
            expected_texts (Optional[List[str]]): Expected texts
            retrieved_texts (Optional[List[str]]): Retrieved texts
            **kwargs: Additional keyword arguments
        """
        if expected_ids is None or retrieved_ids is None:
            raise NotImplementedError("Scoring by Id is not currently implemented.")
        expected_texts = expected_texts or []
        retrieved_texts = retrieved_texts or []
        scorer = rouge_scorer.RougeScorer(["rougeL"], use_stemmer=True)
        total: float = 0.0
        metadata: T.Dict[str, T.Dict[str, T.Dict[str, str | float]]] = {"hits": {}}
        for expected_text in expected_texts:
            scores = [
                (text, scorer.score(text, expected_text)["rougeL"].precision)
                for text in retrieved_texts
            ]
            best_text, best_score = max(scores, key=lambda x: x[1])
            total += best_score
            metadata["hits"][expected_text] = {
                "best_text": best_text,
                "best_score": best_score,
            }
        score = total / len(expected_texts)

        result = RetrievalMetricResult(score=score, metadata=metadata)
        return result


def acc_confidence(accuracy: float, n_samples: int, zscore: float) -> float:
    if n_samples == 0:
        return np.nan
    return zscore * np.sqrt(accuracy * (1 - accuracy) / n_samples)


def lognormal_confidence(values: T.List[float], zscore: float) -> float:
    if len(values) == 0:
        return np.nan
    return zscore * float(np.std(values, ddof=1)) / np.sqrt(len(values))



import pytest
from llama_index.core.callbacks import CBEventType


@pytest.mark.xfail(reason="Will be decommissioned")
def test_react_rag_agent_simple(
    react_agent_rag_flow, correctness_evaluator, llama_debug
):
    question = "Who is Adam Collis?"
    react_agent_rag_flow.agent.chat(question)
    assert llama_debug.get_event_pairs(CBEventType.AGENT_STEP)


@pytest.mark.xfail(reason="Will be decommissioned")
def test_react_rag_agent_system_prompt(
    react_agent_rag_flow_system_prompt, correctness_evaluator, llama_debug
):
    question = "Who is Adam Collis?"
    react_agent_rag_flow_system_prompt.agent.chat(question)
    assert llama_debug.get_event_pairs(CBEventType.AGENT_STEP)



import pytest

from syftr.tuner.core import get_flow_name


@pytest.mark.parametrize(
    "rag_mode, expected",
    [
        ("no_rag", "Flow"),
        ("rag", "RAGFlow"),
        ("react_rag_agent", "ReActAgentFlow"),
        ("critique_rag_agent", "CritiqueAgentFlow"),
        ("sub_question_rag", "SubQuestionRAGFlow"),
    ],
)
def test_flow_name(rag_mode, expected):
    flow_name = get_flow_name(rag_mode)
    assert flow_name == expected, "Bad flow name"



import argparse
import asyncio
import json
import os
import typing as T

import litellm
import llama_index.core.instrumentation as instrument
import optuna
import ray
from aiolimiter import AsyncLimiter
from llama_index.core.evaluation import CorrectnessEvaluator
from llama_index.core.llms.function_calling import FunctionCallingLLM
from openai import AzureOpenAI
from opto.optimizers import OptoPrime
from opto.trace.bundle import bundle
from opto.trace.nodes import ParameterNode

from syftr import flows
from syftr.configuration import cfg
from syftr.core import QAPair
from syftr.evaluation import evaluation
from syftr.llm import get_llm
from syftr.logger import logger
from syftr.optuna_helper import get_pareto_df
from syftr.ray.utils import ray_init
from syftr.studies import StudyConfig, get_default_study_name
from syftr.tuner.qa_tuner import build_flow, eval_dataset

dispatcher = instrument.get_dispatcher()
CorrectnessEvaluator.evaluate = dispatcher.span(CorrectnessEvaluator.evaluate)  # type: ignore
CorrectnessEvaluator.aevaluate = dispatcher.span(CorrectnessEvaluator.aevaluate)  # type: ignore

EVAL_LLM = "gpt-4o-mini"
OPTIMIZER_LLM = "gpt-4o-std"
PARAMS_TO_OPTMIZE = [
    "template",
    "dataset_description",
]


async def quick_eval(
    flow: flows.Flow,
    eval_llm: FunctionCallingLLM,
    dataset: T.List[QAPair],
    rate_limiter: AsyncLimiter,
) -> T.Tuple[float, str]:
    """Evaluates selected flow using eval_llm on a dataset of choice.
    A lightweight evaluation function to be used within the prompt optimization loop.
    """
    evaluator = CorrectnessEvaluator(eval_llm)
    raw_responses = await asyncio.gather(
        *[evaluation.agenerate_pair(qa_pair, flow, rate_limiter) for qa_pair in dataset]
    )
    responses = [raw_response[0] for raw_response in raw_responses]
    eval_results = await asyncio.gather(
        *[
            evaluation.aevaluate_pair(
                qa_pair=qa_pair,
                response=response,  # type: ignore
                evaluator=evaluator,
                rate_limiter=rate_limiter,
            )
            for qa_pair, response in zip(dataset, responses)
            if response
        ]
    )
    results = [
        int(res[0].passing)  # type: ignore
        for res in eval_results
        if res[0].passing is not None  # type: ignore
    ]
    evals = [res[0].feedback for res in eval_results if not res[0].passing]  # type: ignore
    if results:
        return sum(results) / len(results), "\n".join(evals)  # type: ignore
    return 0.0, "The evluation has failed, the results are incorrect."


class TracedFlow:
    def __init__(self, flow):
        object.__setattr__(self, "_flow", flow)
        self.template = ParameterNode(
            self._flow.template,
            description="A prompt for Q&A bot with instructions for complete and accurate answers",
        )
        self.dataset_description = ParameterNode(
            self._flow.dataset_description,
            description="A description of the dataset with the grounding data",
        )

    def __getattr__(self, name):
        return getattr(self._flow, name)


def optimize_prompt(
    flow: flows.Flow,
    optimizer_llm: str,
    eval_llm: str,
    train: T.List[QAPair],
    test: T.List[QAPair],
    rate_limiter: AsyncLimiter,
    num_epochs: int = 5,
) -> flows.Flow:
    """The main prompt optimization function for a flow.
    Uses Trace library with an optimizer LLM to get better prompts using train and test datasets.
    """
    llm = get_llm(optimizer_llm)
    assert isinstance(llm, AzureOpenAI), (
        "Prompt optimization only supports Azure OpenAI"
    )
    evaluator_llm = get_llm(eval_llm)
    logger.info("Evaluating pareto flow on test dataset before optimization...")
    pre_test_acc, _ = asyncio.run(quick_eval(flow, evaluator_llm, test, rate_limiter))
    logger.info("Pre-optimization accuracy on test: %f", pre_test_acc)
    os.environ["AZURE_API_KEY"] = llm.api_key
    os.environ["AZURE_API_BASE"] = llm.azure_endpoint  # type: ignore
    os.environ["AZURE_API_VERSION"] = llm.api_version  # type: ignore
    litellm_model = f"azure/{llm.model}"  # type: ignore
    os.environ["TRACE_LITELLM_MODEL"] = litellm_model

    @bundle()
    def merge_nodes(*args):
        """This function connects Trace nodes that are specified in args.
        For some reason Trace does not catch the usage of annotated templates inside
        our evaluation function and this function takes precomputed accuracy and connects
        specified nodes for optmization.
        """
        for arg in args:
            _ = arg
        nonlocal curr_accuracy
        return curr_accuracy

    tflow = TracedFlow(flow)
    existing_flow_attrs = [arg for arg in PARAMS_TO_OPTMIZE if hasattr(flow, arg)]
    opt_args = [getattr(tflow, arg) for arg in existing_flow_attrs]
    optimizer = OptoPrime(opt_args)
    param_results = []
    for n_epoch in range(1, num_epochs + 1):
        logger.info("Starting optimization epoch %d", n_epoch)
        curr_accuracy, evals = asyncio.run(
            quick_eval(flow, evaluator_llm, train, rate_limiter)
        )
        output = merge_nodes(*opt_args)
        logger.info("Accuracy on epoch %d: %f", n_epoch, curr_accuracy)
        raw_summary = litellm.completion(
            model=litellm_model,
            messages=[
                {
                    "content": f"You are presented with a feedback from a question answering session. Summarize it briefly: {evals}",
                    "role": "user",
                }
            ],
        )
        eval_summary = raw_summary.choices[0].message.content
        feedback = (
            f"The accuracy of question answering session is {curr_accuracy}."
            f"Evaluation summary of question answering session is '{eval_summary}'"
            f"Modify the prompts to help the LLM produce more accurate answers to the provided questions."
            f"Make sure template arguments are in place."
        )
        optimizer.zero_feedback()
        optimizer.backward(output, feedback)
        try:
            logger.info("Generating a new prompt on epoch %s", n_epoch)
            optimizer.step(verbose=True)
        except litellm.exceptions.ContentPolicyViolationError:
            logger.exception("Prompt optimizer hit content policy violation error")
            continue

        for attr in existing_flow_attrs:
            setattr(flow, attr, getattr(tflow, attr).data)

        param_results.append((curr_accuracy, flow))

    _, argmax_flow = max(param_results, key=lambda x: x[0])
    logger.info("Evaluating pareto flow on test dataset after optimization...")

    post_test_acc, _ = asyncio.run(
        quick_eval(argmax_flow, evaluator_llm, test, rate_limiter)
    )
    logger.info("Post-optimiation accuracy on test: %f", post_test_acc)
    return argmax_flow


def opt_flow(
    flow_params: T.Dict[str, T.Any],
    study_config: StudyConfig,
    new_study: optuna.Study,
    parent_number: int,
):
    """Prompt optimization workflow function.
    Gets a pareto flow specified by flow_params, runs prompt optimization and saves it
    in a new study.
    """
    logger.info("Loading training and test datasets for %s", study_config.name)
    train = list(study_config.dataset.iter_examples(partition="train"))
    train = train[: study_config.optimization.num_prompt_optimization_batch]
    test = list(study_config.dataset.iter_examples(partition="test"))
    test = test[: study_config.optimization.num_prompt_optimization_batch]
    logger.info(
        "Training subset size: %d, test subset size: %s",
        len(train),
        len(test),
    )
    rate_limiter = AsyncLimiter(
        study_config.optimization.rate_limiter_max_coros,
        study_config.optimization.rate_limiter_period,
    )
    logger.info("Building pareto flow from params: %s", flow_params)
    try:
        flow = build_flow(study_config=study_config, params=flow_params)
    except Exception:
        logger.exception("Failed to build flow from %s", flow_params)
        return
    logger.info("Optimizing the prompt...")
    try:
        flow = optimize_prompt(
            flow, OPTIMIZER_LLM, EVAL_LLM, train, test, rate_limiter, num_epochs=2
        )
    except Exception:
        logger.exception("Failed to optimize prompt for flow %s", flow)
        return
    logger.info("Resulting optimized flow: %s", flow)
    logger.info("Evaluating updated flow and attaching it to the trial...")
    try:
        eval_results = eval_dataset(
            study_config, study_config.dataset, flow, study_config.evaluation.mode
        )
    except Exception:
        logger.exception("Failed to evaluate resulting flow %s", flow)
        return
    obj1 = eval_results[study_config.optimization.objective_1_name]
    obj2 = eval_results[study_config.optimization.objective_2_name]
    distributions = study_config.search_space.build_distributions(params=flow_params)
    logger.info("Saving optimized prompt...")
    if "enforce_full_evaluation" in flow_params:
        flow_params.pop("enforce_full_evaluation")
    new_trial = optuna.create_trial(
        values=[obj1, obj2],
        params=flow_params,
        distributions=distributions,
        state=optuna.trial.TrialState.COMPLETE,
    )
    new_trial.set_user_attr("parent_number", parent_number)
    for param in PARAMS_TO_OPTMIZE:
        new_trial.set_user_attr(f"optimized_{param}", getattr(flow, param))
    new_study.add_trial(new_trial)
    logger.info("Done optimizing prompt for %s", flow_params)


@ray.remote
def opt_flow_remote(
    flow_params: T.Dict[str, T.Any],
    study_config: StudyConfig,
    new_study: optuna.Study,
    parent_number: int,
):
    """A wrapper for running prompt optimization on a Ray cluster."""
    return opt_flow(flow_params, study_config, new_study, parent_number)


def run_pareto_flows_prompt_optimization(study_path: str, remote: bool = False):
    """
    Main prompt optimzation workflow function.

    Reads study config from a path and initiates a remote or local mode of prompt optmization.
    """
    logger.info("Loading study config from file %s", study_path)
    study_config = StudyConfig.from_file(study_path)
    logger.info("Loading pareto flows for study %s", study_config.name)
    storage = cfg.database.get_optuna_storage()
    study = optuna.load_study(
        study_name=study_config.name,
        storage=storage,
    )
    df_pareto = get_pareto_df(study_config, success_rate=0.5)
    new_study_name = f"{study_config.name}_prompt_optimization"
    logger.info("Creating new study %s for optimized prompts...", new_study_name)
    po_study = optuna.create_study(
        study_name=new_study_name,
        directions=study.directions,
        storage=storage,
        load_if_exists=True,
    )
    logger.info("Running prompt optimization...")
    if remote:
        opts: T.List[T.Any] = []
        for _, row in df_pareto.iterrows():
            flow_params = json.loads(row["user_attrs_flow"])
            parent_number = row["number"]
            opts.append(
                opt_flow_remote.remote(
                    flow_params, study_config, po_study, parent_number
                )
            )
        ray.get(opts)
    else:
        for _, row in df_pareto.iterrows():
            parent_number = row["number"]
            flow_params = json.loads(row["user_attrs_flow"])
            opt_flow(flow_params, study_config, po_study, parent_number)
    logger.info("Done")


def run_opt(study_path: str, remote: bool):
    """Main workflow wrapper."""
    if remote:
        ray_init()
    run_pareto_flows_prompt_optimization(study_path, remote=remote)


def main():
    """An argument parser."""
    parser = argparse.ArgumentParser()
    default = get_default_study_name()
    parser.add_argument(
        "--study-config",
        help=f"Path to study config yaml (default: {default})",
        default=default,
    )
    parser.add_argument(
        "--remote",
        help="Run the optimization remotely",
        default=False,
        action="store_true",
    )
    args = parser.parse_args()
    run_opt(
        study_path=args.study_config,
        remote=args.remote,
    )


if __name__ == "__main__":
    main()



import tempfile
from pathlib import Path

from optuna.storages import RDBStorage

from syftr.configuration import cfg
from syftr.optimization import initialize_from_study
from syftr.optuna_helper import get_completed_trials, get_pareto_df
from syftr.studies import StudyConfig, get_subspace


def test_study_initialization():
    success_rate = 0.9

    src_name = "test-study-src"
    dst_name = "test-study-dst"

    src_file = Path(cfg.paths.test_studies_dir / f"{src_name}.yaml")
    src_config = StudyConfig.from_file(src_file)
    dst_config = src_config.model_copy()
    dst_config.name = dst_name
    dst_config.reuse_study = False
    dst_config.recreate_study = True
    dst_config.optimization.skip_existing = False

    src_storage = RDBStorage(f"sqlite:////{cfg.paths.test_data_dir}/syftr.db")
    df_pareto = get_pareto_df(src_config.name, success_rate, storage=src_storage)
    df_pareto.reset_index(drop=True, inplace=True)
    pareto_space = get_subspace(df_pareto, dst_config.search_space)

    dst_config.search_space = pareto_space

    with tempfile.TemporaryDirectory() as tmpdirname:
        dst_storage = RDBStorage(f"sqlite:///{tmpdirname}/syftr.db")

        initialize_from_study(
            src_config=src_config,
            dst_config=dst_config,
            src_storage=src_storage,
            dst_storage=dst_storage,
            src_df=df_pareto,
        )

        df_dst = get_completed_trials(
            dst_config.name, storage=dst_storage, success_rate=success_rate
        )

    assert len(df_pareto) >= len(df_dst)

    cols = [
        c
        for c in df_pareto.columns
        if c.startswith("params_") or c.startswith("values_")
    ]
    assert df_pareto[cols].equals(df_pareto[cols])



import typing as T

import pytest
from llama_index.core import PromptTemplate, VectorStoreIndex
from llama_index.core.llms.function_calling import FunctionCallingLLM
from llama_index.core.node_parser import SentenceSplitter
from llama_index.core.node_parser.interface import NodeParser
from llama_index.core.storage.docstore.types import BaseDocumentStore

from syftr.agent_flows import LlamaIndexReactRAGAgentFlow
from syftr.flows import (
    CoAAgentFlow,
    CritiqueAgentFlow,
    Flow,
    RAGFlow,
    ReActAgentFlow,
    SubQuestionRAGFlow,
)
from syftr.llm import LLM_NAMES, get_llm, is_function_calling
from syftr.retrievers.build import _build_dense_index, _build_sparse_index
from syftr.storage import SyftrQADataset
from syftr.studies import StudyConfig
from syftr.templates import get_agent_template, get_template
from syftr.tuner.core import build_splitter
from syftr.tuner.qa_tuner import _get_example_retriever

LLMs = {name: get_llm(name) for name in LLM_NAMES}


@pytest.fixture
def tiny_sparse_flow(tiny_sparse_retriever, gpt_4o_mini, rag_template) -> RAGFlow:
    llm, _ = gpt_4o_mini
    return RAGFlow(
        retriever=tiny_sparse_retriever,
        response_synthesizer_llm=llm,
        template=rag_template,
        params={},
    )


@pytest.fixture
def real_sparse_flow(
    real_sparse_retriever, gpt_4o_mini, rag_template
) -> T.Tuple[RAGFlow, StudyConfig]:
    llm, _ = gpt_4o_mini
    retriever, docstore, study_config = real_sparse_retriever
    return RAGFlow(
        retriever=retriever,
        docstore=docstore,
        response_synthesizer_llm=llm,
        template=rag_template,
        additional_context_num_nodes=2,
        params={},
    ), study_config


@pytest.fixture
def tiny_dense_flow(tiny_dense_retriever, gpt_4o_mini, rag_template) -> RAGFlow:
    llm, _ = gpt_4o_mini
    return RAGFlow(
        retriever=tiny_dense_retriever,
        response_synthesizer_llm=llm,
        template=rag_template,
        params={},
    )


@pytest.fixture
def real_dense_flow(
    real_dense_retriever, gpt_4o_mini, rag_template
) -> T.Tuple[RAGFlow, StudyConfig]:
    llm, _ = gpt_4o_mini
    retriever, docstore, study_config = real_dense_retriever
    return RAGFlow(
        retriever=retriever,
        docstore=docstore,
        additional_context_num_nodes=2,
        response_synthesizer_llm=llm,
        template=rag_template,
        params={},
    ), study_config


@pytest.fixture
def real_sparse_flow_few_shot(
    real_sparse_retriever, gpt_4o_mini, few_shot_rag_template, bge_small_no_hf
) -> T.Tuple[RAGFlow, StudyConfig]:
    llm, _ = gpt_4o_mini
    retriever, docstore, study_config = real_sparse_retriever
    params = {"few_shot_top_k": 5}
    get_qa_examples = _get_example_retriever(params, study_config, bge_small_no_hf)
    return RAGFlow(
        retriever=retriever,
        docstore=docstore,
        additional_context_num_nodes=2,
        response_synthesizer_llm=llm,
        template=few_shot_rag_template,
        get_examples=get_qa_examples,
    ), study_config


@pytest.fixture
def real_hybrid_flow(
    real_hybrid_retriever, gpt_4o_mini, rag_template
) -> T.Tuple[RAGFlow, StudyConfig]:
    llm, _ = gpt_4o_mini
    retriever, docstore, study_config = real_hybrid_retriever
    return RAGFlow(
        retriever=retriever,
        docstore=docstore,
        additional_context_num_nodes=2,
        response_synthesizer_llm=llm,
        template=rag_template,
    ), study_config


@pytest.fixture(
    scope="session",
    params=list(
        [
            (name, llm)
            for name, llm in LLMs.items()
            if name
            not in [
                # Going the way of the dodo
                "gpt-35-turbo",
                # Don't have pricing info for these models yet
                "cerebras-llama-31-8B",
                "cerebras-llama-33-70B",
                "gemini-flash-think-exp",
                # Has to be manually turned on
                "datarobot-r1-70B-functions",
                "datarobot-r1-70B-reasoning",
                # Temporarily disabled for testing
                "o1",
                "o3-mini",
                "together-r1",
                "together-V3",
                # Flaky in agent flows
                "llama-33-70B",
                "phi-4",
            ]
        ]
    ),
    ids=list(
        [
            name
            for name in LLMs.keys()
            if name
            not in [
                # Going the way of the dodo
                "gpt-35-turbo",
                # Don't have pricing info for these models yet
                "cerebras-llama-31-8B",
                "cerebras-llama-33-70B",
                "gemini-flash-think-exp",
                # Has to be manually turned on
                "datarobot-r1-70B-functions",
                "datarobot-r1-70B-reasoning",
                # Temporarily disabled for testing
                "o1",
                "o3-mini",
                "together-r1",
                "together-V3",
                # Flaky in agent flows
                "llama-33-70B",
                "phi-4",
            ]
        ]
    ),
)
def llm(request) -> T.Tuple[str, FunctionCallingLLM]:
    name, llm = request.param
    return name, llm


@pytest.fixture(
    scope="session",
    params=list(
        [(name, llm) for name, llm in LLMs.items() if is_function_calling(llm)]
    ),
    ids=list([name for name, llm in LLMs.items() if is_function_calling(llm)]),
)
def critique_llm(request) -> T.Tuple[str, FunctionCallingLLM]:
    name, llm = request.param
    return name, llm


@pytest.fixture(scope="session")
def basic_flow_all_llms(llm):
    _, llm = llm
    return Flow(response_synthesizer_llm=llm, template=get_template("default"))


@pytest.fixture
def real_subquestion_flow(
    llm, real_dense_retriever, rag_template
) -> T.Tuple[SubQuestionRAGFlow, StudyConfig]:
    _, llm = llm
    retriever, docstore, study_config = real_dense_retriever
    return SubQuestionRAGFlow(
        subquestion_engine_llm=llm,
        subquestion_response_synthesizer_llm=llm,
        retriever=retriever,
        docstore=docstore,
        additional_context_num_nodes=2,
        response_synthesizer_llm=llm,
        template=rag_template,
        dataset_name=study_config.dataset.name,
        dataset_description=study_config.dataset.description,
    ), study_config


@pytest.fixture
def react_agent_flow(
    llm, real_dense_retriever, rag_template
) -> T.Tuple[ReActAgentFlow, StudyConfig]:
    _, llm = llm
    retriever, docstore, study_config = real_dense_retriever
    return ReActAgentFlow(
        retriever=retriever,
        docstore=docstore,
        additional_context_num_nodes=2,
        response_synthesizer_llm=llm,
        subquestion_engine_llm=llm,
        subquestion_response_synthesizer_llm=llm,
        template=rag_template,
        dataset_name=study_config.dataset.name,
        dataset_description=study_config.dataset.description,
    ), study_config


@pytest.fixture
def critique_agent_flow(
    critique_llm, real_dense_retriever, rag_template
) -> T.Tuple[CritiqueAgentFlow, StudyConfig]:
    _, llm = critique_llm
    retriever, docstore, study_config = real_dense_retriever
    return CritiqueAgentFlow(
        retriever=retriever,
        docstore=docstore,
        additional_context_num_nodes=2,
        response_synthesizer_llm=llm,
        subquestion_engine_llm=llm,
        subquestion_response_synthesizer_llm=llm,
        critique_agent_llm=llm,
        reflection_agent_llm=llm,
        template=rag_template,
        dataset_name=study_config.dataset.name,
        dataset_description=study_config.dataset.description,
    ), study_config


@pytest.fixture
def react_agent_flow_hybrid_hyde_reranker_few_shot(
    real_hybrid_retriever, gpt_4o_mini, few_shot_rag_template, bge_small_no_hf
) -> T.Tuple[ReActAgentFlow, StudyConfig]:
    llm, _ = gpt_4o_mini
    retriever, docstore, study_config = real_hybrid_retriever
    params = {"few_shot_top_k": 5}
    get_qa_examples = _get_example_retriever(params, study_config, bge_small_no_hf)
    return ReActAgentFlow(
        retriever=retriever,
        docstore=docstore,
        additional_context_num_nodes=2,
        response_synthesizer_llm=llm,
        subquestion_engine_llm=llm,
        subquestion_response_synthesizer_llm=llm,
        template=few_shot_rag_template,
        get_examples=get_qa_examples,
        dataset_name=study_config.dataset.name,
        dataset_description=study_config.dataset.description,
        hyde_llm=llm,
        reranker_llm=llm,
        reranker_top_k=5,
    ), study_config


@pytest.fixture
def coa_agent_flow(
    real_sparse_retriever, gpt_4o_mini, rag_template
) -> T.Tuple[CoAAgentFlow, StudyConfig]:
    llm, _ = gpt_4o_mini
    retriever, docstore, study_config = real_sparse_retriever
    return CoAAgentFlow(
        retriever=retriever,
        docstore=docstore,
        response_synthesizer_llm=llm,
        template=rag_template,
        dataset_name=study_config.dataset.name,
        dataset_description=study_config.dataset.description,
        enable_calculator=True,
    ), study_config


@pytest.fixture(scope="session")
def cot_template():
    return get_template("CoT", with_context=False)


@pytest.fixture(scope="session")
def cot_template_rag():
    return get_template("CoT", with_context=True)


@pytest.fixture(scope="session")
def sentence_splitter() -> SentenceSplitter:
    return SentenceSplitter(
        chunk_size=256,
        chunk_overlap=64,
    )


@pytest.fixture(scope="session")
def recursive_splitter(bge_small_no_hf, real_dataset_study_config) -> NodeParser:
    return build_splitter(
        real_dataset_study_config,
        {
            "splitter_method": "recursive",
            "splitter_chunk_exp": 10,
            "splitter_chunk_overlap_frac": 0.5,
        },
    )


@pytest.fixture(scope="session")
def simple_dense_index(
    tiny_dataset, sentence_splitter, bge_small_no_hf
) -> T.Tuple[VectorStoreIndex, BaseDocumentStore]:
    return _build_dense_index(
        documents=list(tiny_dataset.iter_grounding_data()),
        transforms=[sentence_splitter],
        embedding_model=bge_small_no_hf,
    )


@pytest.fixture(scope="session")
def real_dense_index(
    real_dataset, sentence_splitter, bge_small_no_hf
) -> T.Tuple[VectorStoreIndex, BaseDocumentStore, SyftrQADataset]:
    index, docstore = _build_dense_index(
        documents=list(real_dataset.iter_grounding_data()),
        transforms=[sentence_splitter],
        embedding_model=bge_small_no_hf,
        max_chunks=100,
    )
    return index, docstore, real_dataset


@pytest.fixture(scope="session")
def hotpot_index(sentence_splitter, bge_small_no_hf):
    hotpot = StudyConfig.from_file(
        "tests/functional/data/studies/test-hotpot-toy.yaml"
    ).dataset
    return (
        _build_dense_index(
            documents=list(hotpot.iter_grounding_data()),
            transforms=[sentence_splitter],
            embedding_model=bge_small_no_hf,
        ),
        hotpot,
    )


@pytest.fixture(scope="session")
def financebench_index(
    sentence_splitter, bge_small_no_hf
) -> T.Tuple[VectorStoreIndex, BaseDocumentStore, SyftrQADataset]:
    financebench = StudyConfig.from_file(
        "tests/functional/data/studies/financebench-pepsi.yaml"
    ).dataset
    index, docstore = _build_dense_index(
        documents=list(financebench.iter_grounding_data()),
        transforms=[sentence_splitter],
        embedding_model=bge_small_no_hf,
        max_chunks=100,
    )
    return index, docstore, financebench


@pytest.fixture(scope="session")
def real_sparse_index(real_dataset, recursive_splitter):
    return _build_sparse_index(
        documents=real_dataset.iter_grounding_data(),
        transforms=[recursive_splitter],
        top_k=5,
    )


@pytest.fixture
def llama_index_flow(
    cot_template_rag,
    simple_dense_index,
    bge_small_no_hf,
) -> RAGFlow:
    """Simple RAG flow."""
    name = "gpt-4o-mini"
    llm = get_llm(name)
    dense_retriever = simple_dense_index.as_retriever(
        embed_model=bge_small_no_hf, similarity_top_k=5
    )
    return RAGFlow(
        retriever=dense_retriever,
        response_synthesizer_llm=llm,
        template=cot_template_rag,
    )


@pytest.fixture
def react_agent_rag_flow(
    hotpot_index, financebench_index
) -> LlamaIndexReactRAGAgentFlow:
    fb_index, _ = financebench_index
    hp_index, _ = hotpot_index
    llm = get_llm("gpt-4o-mini")
    return LlamaIndexReactRAGAgentFlow(
        indexes=[
            ("hotpot", "A collection of trivia from Wikipedia", hp_index),
            (
                "financebench",
                "Finacial reports and accounting data",
                fb_index,
            ),
        ],
        llm=llm,
        template=get_template("default", with_context=True),
        verbose=True,
    )  # type: ignore


@pytest.fixture
def react_agent_rag_flow_system_prompt(
    hotpot_index, financebench_index
) -> LlamaIndexReactRAGAgentFlow:
    fb_index, _ = financebench_index
    hp_index, _ = hotpot_index
    llm = get_llm("gpt-4o-mini")
    prompt = get_agent_template("aggressive")
    return LlamaIndexReactRAGAgentFlow(
        indexes=[
            ("hotpot", "A collection of trivia from Wikipedia", hp_index),
            (
                "financebench",
                "Finacial reports and accounting data",
                fb_index,
            ),
        ],
        llm=llm,
        system_prompt=PromptTemplate(prompt),
        template=get_template("default", with_context=True),
        verbose=True,
    )  # type: ignore



from llama_index.core.evaluation import EvaluationResult

from syftr.evaluation.evaluator_factory import json_parser_function


def parse_correctness_evaluation(
    query: str, response: str, threshold: float = 4.0
) -> EvaluationResult:
    score, feedback = json_parser_function(response)
    if score is None:
        raise ValueError(f"Could not parse score from response: {response}")
    passing = score >= threshold
    return EvaluationResult(
        query=query, response=response, passing=passing, score=score, feedback=feedback
    )


def parse_correctness_evaluation_ten(
    query: str, response: str, threshold: float = 8.0
) -> EvaluationResult:
    return parse_correctness_evaluation(query, response, threshold=threshold)


def parse_correctness_evaluation_comparison(
    query: str, response: str, threshold: float = 2.0
) -> EvaluationResult:
    return parse_correctness_evaluation(query, response, threshold=threshold)


def parse_correctness_evaluation_simple(
    query: str, response: str, threshold: float = 0.0
) -> EvaluationResult:
    response_formatted = response.upper().replace(" ", "")
    passing = "*YES*" in response_formatted and "*NO*" not in response_formatted
    feedback = response
    return EvaluationResult(
        query=query, response=response, passing=passing, feedback=feedback
    )



import getpass
import os

import pytest

from syftr.huggingface_helper import load_hf_token_into_env
from syftr.startup import download_nltk_data

download_nltk_data()
load_hf_token_into_env()


def pytest_addoption(parser):
    parser.addoption("--gpu", action="store_true", help="Use GPU for embeddings")


@pytest.fixture(scope="session", autouse=True)
def set_ray_tmpdir():
    username = getpass.getuser()
    os.environ["RAY_TMPDIR"] = f"/tmp/ray_test_{username}"
    yield



from pathlib import Path

from syftr.configuration import cfg
from syftr.studies import StudyConfig
from tests.check_trials import check_trials_are_flawless
from tests.e2e.optimization import run_and_test_optimization


def test_seeding_and_optimization_using_lats():
    study_config_file_name = "test-lats.yaml"
    study_config_file = Path(cfg.paths.test_studies_dir / study_config_file_name)
    study_config = StudyConfig.from_file(study_config_file)
    study_config.optimization.use_hf_embedding_models = False

    df_trials = run_and_test_optimization(study_config)
    check_trials_are_flawless(df_trials)



import concurrent
import concurrent.futures
import json
import random
import typing as T

import optuna
import pandas as pd
import ray
from optuna.storages import BaseStorage
from ray import tune
from ray._private.worker import RemoteFunction1

from syftr.configuration import cfg
from syftr.logger import logger
from syftr.optuna_helper import (
    get_completed_trials,
    get_knee_flow,
    get_sampler,
    recreate_with_completed_trials,
    seed_study,
    trial_exists,
    without_non_search_space_params,
)
from syftr.studies import StudyConfig


def tune_objective(
    objective,
    study_config: StudyConfig,
    components: T.List[str],
    block_name: str | None = None,
) -> T.Dict[str, float] | None:
    study = optuna.load_study(
        study_name=study_config.name,
        storage=cfg.database.get_optuna_storage(),
    )
    trial = study.ask()
    if block_name:
        trial.set_user_attr("block_name", block_name)
    try:
        obj_1, obj_2 = objective(trial, study_config, components)
        study.tell(trial, [obj_1, obj_2])
        return {
            study_config.optimization.objective_1_name: obj_1,
            study_config.optimization.objective_2_name: obj_2,
        }
    except optuna.TrialPruned:
        return None


def user_confirm_delete(study_config: StudyConfig) -> bool:
    study_name = study_config.name

    if cfg.optuna.noconfirm:
        logger.warning(
            "noconfirm set; we are allowed to delete study %s if it exists", study_name
        )
        return True

    confirm = input(
        f"Are you sure you want to overwrite study {study_name} if it exists? yes/no\n>>> "
    )
    if confirm != "yes":
        logger.warning(f"Cowardly refusing to delete study {study_name}")
        return False
    return True


def save_study_config_to_db(study: optuna.Study, study_config: StudyConfig):
    attrs = study_config.model_dump(mode="json")
    logger.info("Saving study config of %s to the database", study.study_name)
    for attr, value in attrs.items():
        study.set_user_attr(attr, value)


def get_study(
    study_config: StudyConfig, storage: BaseStorage | None = None
) -> optuna.Study:
    study_name = study_config.name
    storage = storage or cfg.database.get_optuna_storage()
    if study_config.reuse_study:
        logger.info(
            "We are going to reuse study '%s' if it exists or create a new one otherwise",
            study_name,
        )
        if study_config.recreate_study:
            recreate_with_completed_trials(study_config, storage)
    elif user_confirm_delete(study_config):
        try:
            optuna.delete_study(study_name=study_name, storage=storage)
            logger.info("Study '%s' deleted", study_name)
        except KeyError:
            logger.info(
                "Study '%s' does not exist and we will create a new one",
                study_name,
            )

    sampler = get_sampler(study_config)
    study = optuna.create_study(
        study_name=study_name,  # heartbeat doesn't work when using ask and tell
        storage=storage,
        load_if_exists=study_config.reuse_study,
        directions=[
            "maximize",
            "minimize",
        ],
        sampler=sampler,
    )
    save_study_config_to_db(study, study_config)
    return study


def _get_user_attrs(row: pd.Series) -> T.Dict[str, T.Any]:
    user_attrs = {}
    for key, value in row.items():
        if key.startswith("user_attrs_"):
            user_attrs[key.replace("user_attrs_", "")] = value
    return user_attrs


def initialize_from_study(
    src_config: StudyConfig,
    dst_config: StudyConfig,
    src_storage: BaseStorage | None = None,
    dst_storage: BaseStorage | None = None,
    success_rate: float | None = None,
    src_df: pd.DataFrame | None = None,
) -> optuna.Study:
    """
    Initialize a study with trials from another study from the same Optuna storage.
    Args:
        src_config: Source study configuration.
        dst_config: Destination study configuration.
        src_storage: Storage for the source study. If not provided, the default storage is used.
        dst_storage: Storage for the destination study. If not provided, the source storage is used.
        storage: Storage for the studies.
        success_rate: Success rate for filtering trials.
        src_df: DataFrame containing trials to initialize from. If no DataFrame is provided, all trials will be fetched from the source study.
    """
    assert success_rate is None or src_df is None, (
        "Cannot use both success_rate and src_df"
    )

    src_storage = src_storage or cfg.database.get_optuna_storage()
    dst_storage = dst_storage or src_storage

    src_name = src_config.name
    dst_name = dst_config.name

    logger.info("Initializing study '%s' from study '%s'", dst_name, src_name)

    study = get_study(dst_config, storage=dst_storage)

    if src_df is None:
        src_df = get_completed_trials(
            src_name, storage=src_storage, success_rate=success_rate
        )
    for _, row in src_df.iterrows():
        if not isinstance(row["user_attrs_flow"], str):
            logger.info(
                "Skipping invalid flow in source study '%s': %s",
                src_name,
                row["user_attrs_flow"],
            )
            continue
        src_flow = json.loads(row["user_attrs_flow"])
        src_flow = without_non_search_space_params(src_flow, src_config)

        if dst_config.optimization.skip_existing and trial_exists(
            dst_name, src_flow, dst_storage
        ):
            logger.warning(
                "Flow already exists in destination study '%s': %s",
                dst_name,
                str(src_flow),
            )
            continue

        distributions = dst_config.search_space.build_distributions(params=src_flow)

        obj1 = row["values_0"]
        obj2 = row["values_1"]

        trial = optuna.create_trial(
            values=[obj1, obj2],
            params=src_flow,
            distributions=distributions,
            user_attrs=_get_user_attrs(row),
        )

        study.add_trial(trial)
        logger.debug("Initialization added trial with params: %s", trial.params)

    return study


class StudyRunner:
    """
    Class to run the study optimization block by block.
    A block contains flow components that are optimized together while the parameter
    values for other components are fixed.
    The optimization is done using Ray Tune, which allows for parallel evaluation of trials.
    For a global optimization, there would be only one block with all components.
    """

    def __init__(
        self, study_config: StudyConfig, objective: T.Callable, seeder: RemoteFunction1
    ):
        self.study_config = study_config
        self.name = study_config.name
        self.objective = objective
        self.seeder = seeder
        self.seeder_timeout = study_config.optimization.seeder_timeout
        self.method = study_config.optimization.method
        self.blocks = study_config.optimization.blocks.copy()
        self.shuffle_blocks = study_config.optimization.shuffle_blocks
        self.num_trials = study_config.optimization.num_trials
        self.cpus_per_trial = self.study_config.optimization.cpus_per_trial
        self.gpus_per_trial = self.study_config.optimization.gpus_per_trial
        self.max_concurrent_trials = (
            self.study_config.optimization.max_concurrent_trials
        )
        self.raise_on_failed_trial = (
            self.study_config.optimization.raise_on_failed_trial
        )
        self.success_rate = self.study_config.optimization.pareto_eval_success_rate

    def run(self) -> optuna.Study:
        """
        Start the optimization process as configured.
        For every block, the components are optimized using Ray Tune.
        We run as many trails for a block as specified in the configuration.
        If after one iteration through all blocks, the maximum number of trials
        has not been reached, we start over with another round.
        This allows for a step-by-step optimization where we run only a few trials
        per block and revisit the blocks.
        Having only a single block is equivalent to a global optimization.
        """
        assert ray.is_initialized(), "Please initialize ray"
        study_config = self.study_config

        study = get_study(study_config)

        with concurrent.futures.ThreadPoolExecutor() as executor:
            future = executor.submit(seed_study, self.seeder, study_config)
            if self.seeder_timeout is not None:
                logger.info("Seeding timeout is %i seconds", int(self.seeder_timeout))
            else:
                logger.info("Seeding without timeout")
            try:
                future.result(timeout=self.seeder_timeout)
            except concurrent.futures.TimeoutError:
                logger.warning("Seeding continues after timeout")
            logger.info("Starting ray tune optimization")

            components = []
            trials = 0
            while trials < self.num_trials:
                if self.shuffle_blocks:
                    logger.info("Shuffling optimization blocks")
                    random.shuffle(self.blocks)
                for block in self.blocks:
                    logger.info("Starting optimization block: %s", block.name)
                    if self.method == "expanding":
                        logger.info(
                            "Expanding searchable components with block: %s", block.name
                        )
                        components.extend(block.components)
                    else:
                        logger.info(
                            "Using searchable components from block: %s", block.name
                        )
                        components = block.components
                    num_trials = min(block.num_trials, self.num_trials - trials)
                    trials += num_trials
                    tune.run(
                        lambda _: tune_objective(
                            study_config=study_config,
                            objective=self.objective,
                            components=components,
                            block_name=block.name,
                        ),
                        num_samples=num_trials,
                        resources_per_trial={
                            "CPU": self.cpus_per_trial,
                            "GPU": self.gpus_per_trial,
                        },
                        max_concurrent_trials=self.max_concurrent_trials,
                        checkpoint_config=ray.train.CheckpointConfig(
                            checkpoint_frequency=0
                        ),
                        raise_on_failed_trial=self.raise_on_failed_trial,
                    )
                    if self.method == "knee":
                        flow = get_knee_flow(
                            study_config=study_config,
                            success_rate=self.success_rate,
                        )
                        defaults = {
                            key: value
                            for key, value in flow.items()
                            if key in components
                        }
                        logger.info(
                            "Knee flow defaults from block '%s': %s",
                            block.name,
                            defaults,
                        )
                        study_config.search_space.update_defaults(defaults)

                future.result()  # Wait for seeding to finish

        logger.info("Finished optimizing '%s'", self.name)
        return study
